"""
Base class for all generators

Developer Notes
---------------

Subclasses of this class implement specific generators. Each generator is in one of two styles:

1. schemaloader-based, using a Visitor pattern (older)
2. schemaview-based, no visitor pattern (newer)

New generators should always using the latter approach

See: https://github.com/linkml/linkml/issues/923

"""

import abc
import logging
import os
import re
import sys
from collections.abc import Mapping
from dataclasses import dataclass, field
from functools import lru_cache
from pathlib import Path
from typing import Callable, ClassVar, Optional, TextIO, Union, cast

import click
from click import Argument, Command, Option
from jsonasobj2 import JsonObj
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    Definition,
    Element,
    ElementName,
    EnumDefinition,
    EnumDefinitionName,
    PrefixPrefixPrefix,
    SchemaDefinition,
    SlotDefinition,
    SlotDefinitionName,
    SubsetDefinition,
    SubsetDefinitionName,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.namespaces import Namespaces

from linkml import LOCAL_METAMODEL_YAML_FILE
from linkml.cli.logging import DEFAULT_LOG_LEVEL_INT, log_level_option
from linkml.utils.mergeutils import alias_root
from linkml.utils.schemaloader import SchemaLoader
from linkml.utils.typereferences import References

logger = logging.getLogger(__name__)


@lru_cache
def _resolved_metamodel(mergeimports):
    if not os.path.exists(LOCAL_METAMODEL_YAML_FILE):
        raise AssertionError(f"{LOCAL_METAMODEL_YAML_FILE} not found")

    base_dir = str(Path(str(LOCAL_METAMODEL_YAML_FILE)).parent)
    logger.debug(f"BASE={base_dir}")
    metamodel = SchemaLoader(
        LOCAL_METAMODEL_YAML_FILE,
        importmap={"linkml": base_dir},
        base_dir=base_dir,
        mergeimports=mergeimports,
    )
    metamodel.resolve()
    return metamodel


@dataclass
class Generator(metaclass=abc.ABCMeta):
    """
    Base class for generators

    For usage `Generator Docs <https://linkml.io/linkml/generators/>`_
    """

    schema: Union[str, TextIO, SchemaDefinition, "Generator", Path]
    """metamodel compliant schema.  Can be URI, file name, actual schema, another generator, an
        open file or a pre-parsed schema"""

    # ClassVars
    generatorname: ClassVar[str] = None
    """ Name of the generator. Override with os.path.basename(__file__)"""

    generatorversion: ClassVar[str] = None  # Generator version identifier
    """Version of the generator. Consider deprecating and instead use overall linkml version"""

    uses_schemaloader: ClassVar[bool] = True
    """Old-style generator that uses the SchemaLoader and visitor pattern"""

    # uses_schemaview: ClassVar[bool] = True
    # """New-style generator that uses SchemaView"""

    requires_metamodel: ClassVar[bool] = True
    """Generator queries an instance of the metamodel"""

    valid_formats: ClassVar[list[str]] = []
    """Allowed formats - first format is default"""

    visit_all_class_slots: ClassVar[bool] = False
    """Visitor ClassVar: False means only visit own slots, True means visit all slots"""

    visits_are_sorted: ClassVar[bool] = True
    """Visitor ClassVar: True means visit basic types in alphabetial order, false in entry"""

    sort_class_slots: ClassVar[bool] = False
    """Visitor ClassVar: True means visit class slots in alphabetical order"""

    # Object-level Vars
    schemaview: Optional[SchemaView] = None
    """A wrapper onto the schema object"""

    format: Optional[str] = None
    """expected output format"""

    file_extension: ClassVar[str] = None

    metadata: bool = True
    """True means include date, generator, etc. information in source header if appropriate"""

    useuris: Optional[bool] = None
    """True means declared class slot uri's are used.  False means use model uris"""

    log_level: Optional[int] = DEFAULT_LOG_LEVEL_INT
    """Logging level, 0 is minimum"""

    mergeimports: Optional[bool] = True
    """True means merge non-linkml sources into importing package.  False means separate packages"""

    source_file_date: Optional[str] = None
    """Modification date of input source file"""

    source_file_size: Optional[int] = None
    """Size of the source file in bytes"""

    logger: Optional[logging.Logger] = None
    """Logger to use for logging messages"""

    verbose: Optional[bool] = None
    """Verbosity"""

    output: Optional[str] = None
    """Path to output file. Note all generators may not implement this
    uniformly, see https://github.com/linkml/linkml/issues/923"""

    namespaces: Optional[Namespaces] = None
    """All prefix expansions used"""

    directory_output: bool = False
    """True means output is to a directory, False is to stdout"""

    base_dir: str = None  # Base directory of schema
    """Working directory or base URL of sources.
    Setting this is necessary for correct retrieval of relative imports."""

    metamodel_name_map: dict[str, str] = None
    """Allows mapping of names of metamodel elements such as slot, etc"""

    importmap: Optional[Union[str, Optional[Mapping[str, str]]]] = None
    """File name of import mapping file -- maps import name/uri to target"""

    emit_prefixes: set[str] = field(default_factory=lambda: set())
    """Prefixes to emit, for RDF-based generators"""

    metamodel: SchemaLoader = None
    """A SchemaLoader instance that points to the LinkML metamodel (meta.yaml)"""

    stacktrace: bool = False
    """True means print stack trace, false just error message"""

    include: Optional[Union[str, Path, SchemaDefinition]] = None
    """If set, include extra schema outside of the imports mechanism"""

    def __post_init__(self) -> None:
        if not self.logger:
            self.logger = logger
        if self.log_level is not None:
            self.logger.setLevel(self.log_level)
        if self.format is None:
            self.format = self.valid_formats[0]
        if self.format not in self.valid_formats:
            raise ValueError(f"Unrecognized format: {format}; known={self.valid_formats}")
        # legacy: all generators should use "mergeimports"
        # self.merge_imports = self.mergeimports
        if not self.metadata:
            self.source_file_date = None
            self.source_file_size = None
        if self.requires_metamodel:
            self.metamodel = _resolved_metamodel(self.mergeimports)

        schema = self.schema
        if isinstance(schema, Path):
            schema = str(schema)

        # TODO: remove aliasing
        self.emit_metadata = self.metadata
        if self.uses_schemaloader:
            self._initialize_using_schemaloader(schema)
        else:
            self.logger.info(f"Using SchemaView with im={self.importmap} // base_dir={self.base_dir}")
            self.schemaview = SchemaView(schema, importmap=self.importmap, base_dir=self.base_dir)
            if self.include:
                if isinstance(self.include, (str, Path)):
                    self.include = SchemaView(self.include, importmap=self.importmap, base_dir=self.base_dir).schema
                self.schemaview.merge_schema(self.include)
            self.schema = self.schemaview.schema

        self._init_namespaces()

    def _initialize_using_schemaloader(self, schema: Union[str, TextIO, SchemaDefinition, "Generator"]):
        # currently generators are very liberal in what they accept, including
        # other generators.
        # See https://github.com/linkml/linkml/issues/923 for discussion on how
        # to simplify the overall framework
        if isinstance(schema, Generator):
            gen = schema
            self.schema = gen.schema
            self.synopsis = gen.synopsis
            self.loaded = gen.loaded
            self.namespaces = gen.namespaces
            self.base_dir = gen.base_dir
            self.importmap = gen.importmap
            self.source_file_data = gen.source_file_date
            self.source_file_size = gen.source_file_size
            self.schema_location = gen.schema_location
            self.schema_defaults = gen.schema_defaults
            self.logger = gen.logger
        else:
            if isinstance(schema, SchemaDefinition):
                # schemaloader based methods require schemas to have been created via SchemaLoader,
                # which prepopulates some fields (e.g. definition_url). If the schema has not been processed through the
                # loader, then roundtrip
                schema = schema._as_dict
            loader = SchemaLoader(
                schema,
                self.base_dir,
                useuris=self.useuris,
                importmap=self.importmap,
                logger=self.logger,
                mergeimports=self.mergeimports,
                emit_metadata=self.metadata,
                source_file_date=self.source_file_date,
                source_file_size=self.source_file_size,
            )
            loader.resolve()
            self.schema = loader.schema
            self.synopsis = loader.synopsis
            self.loaded = loader.loaded
            self.namespaces = loader.namespaces
            self.base_dir = loader.base_dir
            self.importmap = loader.importmap
            self.source_file_data = loader.source_file_date
            self.source_file_size = loader.source_file_size
            self.schema_location = loader.schema_location
            self.schema_defaults = loader.schema_defaults

    def _init_namespaces(self):
        if self.namespaces is None:
            self.namespaces = Namespaces()
            if isinstance(self.schema.prefixes, dict):
                for key, value in self.schema.prefixes.items():
                    self.namespaces[key] = value
            elif isinstance(self.schema.prefixes, JsonObj):
                prefixes = vars(self.schema.prefixes)
                for key, value in prefixes.items():
                    self.namespaces[key] = value
            else:
                for prefix in self.schema.prefixes.values():
                    self.namespaces[prefix.prefix_prefix] = prefix.prefix_reference

    def serialize(self, **kwargs) -> str:
        """
        Generate output in the required format

        :param kwargs: Generator specific parameters
        :return: Generated output
        """
        out = ""

        # the default is to use the Visitor Pattern; each individual generator may
        # choose to override methods {visit,end}_{element}.
        # See https://github.com/linkml/linkml/issues/923
        sub_out = self.visit_schema(**kwargs)
        if sub_out is not None:
            out += sub_out
        for sn, ss in (
            sorted(self.schema.subsets.items(), key=lambda s: s[0].lower())
            if self.visits_are_sorted
            else self.schema.subsets.items()
        ):
            sub_out = self.visit_subset(ss)
            if sub_out is not None:
                out += sub_out
        for tn, typ in (
            sorted(self.schema.types.items(), key=lambda s: s[0].lower())
            if self.visits_are_sorted
            else self.schema.types.items()
        ):
            sub_out = self.visit_type(typ)
            if sub_out is not None:
                out += sub_out
        for enum in (
            sorted(self.schema.enums.values(), key=lambda e: e.name.lower())
            if self.visits_are_sorted
            else self.schema.enums.values()
        ):
            sub_out = self.visit_enum(enum)
            if sub_out is not None:
                out += sub_out
        for sn, slot in (
            sorted(self.schema.slots.items(), key=lambda c: c[0].lower())
            if self.visits_are_sorted
            else self.schema.slots.items()
        ):
            sub_out = self.visit_slot(self.aliased_slot_name(slot), slot)
            if sub_out is not None:
                out += sub_out
        for cls in (
            sorted(self.schema.classes.values(), key=lambda c: c.name.lower())
            if self.visits_are_sorted
            else self.schema.classes.values()
        ):
            cls_out = self.visit_class(cls)
            if cls_out:
                if isinstance(cls_out, str):
                    out += cls_out
                for slot in self.all_slots(cls) if self.visit_all_class_slots else self.own_slots(cls):
                    sub_out = self.visit_class_slot(cls, self.aliased_slot_name(slot), slot)
                    if sub_out is not None:
                        out += sub_out
                sub_out = self.end_class(cls)
                if sub_out is not None:
                    out += sub_out
        sub_out = self.end_schema(**kwargs)
        if sub_out is not None:
            out += sub_out
        out = out.rstrip()
        if not (out.count("\n") == 0 and out.startswith("http")):
            out += "\n"
        return out

    def visit_schema(self, **kwargs) -> Optional[str]:
        """Visited once at the beginning of generation

        @param kwargs: Arguments passed through from CLI -- implementation dependent
        """
        ...

    def end_schema(self, **kwargs) -> Optional[str]:
        """Visited once at the end of generation

        @param kwargs: Arguments passed through from CLI -- implementation dependent
        """
        ...

    def visit_class(self, cls: ClassDefinition) -> Optional[Union[str, bool]]:
        """Visited once per schema class

        @param cls: class being visited
        @return: Visit slots and end class.  False means skip and go on
        """
        return True

    def end_class(self, cls: ClassDefinition) -> Optional[str]:
        """Visited after visit_class_slots (if visit_class returned true)

        @param cls: class being visited
        """
        ...

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> Optional[str]:
        """Visited for each slot in a class.  If class level visit_all_slots is true, this is visited once
        for any class that is inherited (class itself, is_a, mixin, apply_to).  Otherwise, just the own slots.

        @param cls: containing class
        @param aliased_slot_name: Aliased slot name.  May not be unique across all class slots
        @param slot: being visited
        """
        ...

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> Optional[str]:
        """Visited once for every slot definition in the schema.

        @param aliased_slot_name: Aliased name of the slot.  May not be unique
        @param slot: visited slot
        """
        ...

    def visit_type(self, typ: TypeDefinition) -> Optional[str]:
        """Visited once for every type definition in the schema

        @param typ: Type definition
        """
        ...

    def visit_subset(self, subset: SubsetDefinition) -> Optional[str]:
        """Visited once for every subset definition in the schema

        #param subset: Subset definition
        """
        ...

    def visit_enum(self, enum: EnumDefinition) -> Optional[str]:
        """Visited once for every enum definition in the schema

        @param enum: Enum definition
        """
        ...

    # =============================
    # Helper methods
    # =============================
    def own_slots(self, cls: Union[ClassDefinitionName, ClassDefinition]) -> list[SlotDefinition]:
        """Return the list of slots owned the class definition.  An "own slot" is any ``cls`` slot that does not appear
        in the class is_a parent.  Own_slots include:

            * any slot whose domain is cls
            * slot_usage entries
            * slots from mixins entries
            * slots from apply_to entries


        @param cls: class name or class definition name
        @return: list of owned slots.  List is sorted if sort_class_slots is true, otherwise in class order
        """
        if not isinstance(cls, ClassDefinition):
            cls = self.schema.classes[cls]
        parent = self.schema.classes[cls.is_a] if cls.is_a else None
        seen = set()
        rval = []
        for sname in cls.slots:
            sname_base = alias_root(self.schema, sname)
            if sname_base not in seen and (not parent or sname not in parent.slots):
                slot = self.schema.slots[sname]
                rval.append(slot)
                seen.add(sname_base)
        return sorted(rval, key=lambda s: s.name) if self.sort_class_slots else rval

    def own_slot_names(self, cls: Union[ClassDefinitionName, ClassDefinition]) -> list[SlotDefinitionName]:
        return [slot.name for slot in self.own_slots(cls)]

    def all_slots(
        self,
        cls: Union[ClassDefinitionName, ClassDefinition],
        *,
        cls_slots_first: bool = False,
        seen: Optional[set[ClassDefinitionName]] = None,
    ) -> list[SlotDefinition]:
        """Return all slots that are part of the class definition.  This includes all is_a, mixin and apply_to slots
        but does NOT include slot_usage targets.  If class B has a slot_usage entry for slot "s", only the slot
        definition for the redefined slot will be included, not its base.  Slots are added in the order they appear
        in classes, with recursive is_a's being added first followed by mixins and finally apply_tos

        @param cls: class definition or class definition name
        @param cls_slots_first: True means return own slots at the top of the list
        @param seen: List of slots already recorded. Used for internal recursion
        @return: ordered list of slots in the class with slot usages removed
        """
        if not isinstance(cls, ClassDefinition):
            cls = self.schema.classes[cls]
        if seen is None:
            seen = set()
        rval = []

        parent = self.schema.classes[cls.is_a] if cls.is_a else None
        if cls_slots_first:
            for slot in self.own_slots(cls):
                sname_base = alias_root(self.schema, slot.name)
                if sname_base not in seen:
                    rval.append(slot)
                    seen.add(sname_base)
            return rval + (self.all_slots(parent, cls_slots_first=cls_slots_first, seen=seen) if parent else [])
        else:
            for sname in cls.slots:
                sname_base = alias_root(self.schema, sname)
                if sname_base not in seen:
                    slot = self.schema.slots[sname]
                    rval.append(slot)
                    seen.add(sname_base)
            return sorted(rval, key=lambda s: s.name) if self.sort_class_slots else rval

    def parent(
        self, element: Union[ClassDefinition, SlotDefinition]
    ) -> Optional[Union[ClassDefinition, SlotDefinition]]:
        """Return the parent of element, if any"""
        return (
            None
            if element.is_a is None
            else (
                self.schema.classes[element.is_a]
                if isinstance(element, ClassDefinition)
                else self.schema.slots[element.is_a]
            )
        )

    def ancestors(self, element: Union[ClassDefinition, SlotDefinition]) -> list[ElementName]:
        """Return an ordered list of ancestor names for the supplied slot or class

        @param element: Slot or class name or definition
        @return: Ordered list of ancestor names
        """
        return [element.name] + ([] if element.is_a is None else self.ancestors(self.parent(element)))

    def neighborhood(self, elements: Union[str, ElementName, list[ElementName]]) -> References:
        """Return a list of all slots, classes and types that touch any element in elements, including the element
        itself

        @param elements: Element names to do proximity with
        @return: All slots and classes that touch element
        """
        if isinstance(elements, (str, ElementName)):
            elements = [elements]
        touches = References()
        for element in elements:
            if element in self.schema.classes:
                touches.classrefs.add(cast(ClassDefinitionName, element))
                cls = self.schema.classes[cast(ClassDefinitionName, element)]
                if cls.is_a:
                    touches.classrefs.add(cls.is_a)
                # Mixins include apply_to's
                touches.classrefs.update(set(cls.mixins))
                for slotname in cls.slots:
                    slot = self.schema.slots[slotname]
                    if slot.range in self.schema.classes:
                        touches.classrefs.add(cast(ClassDefinitionName, slot.range))
                    elif slot.range in self.schema.types:
                        touches.typerefs.add(cast(TypeDefinitionName, slot.range))
                for cv in self.schema.classes.values():
                    if cv.is_a == element:
                        touches.classrefs.add(cv.name)
                if element in self.synopsis.rangerefs:
                    for slotname in self.synopsis.rangerefs[element]:
                        touches.slotrefs.add(slotname)
                        if self.schema.slots[slotname].domain:
                            touches.classrefs.add(self.schema.slots[slotname].domain)
                if cls.in_subset:
                    touches.subsetrefs.update(cls.in_subset)
            if element in self.schema.slots:
                touches.slotrefs.add(cast(SlotDefinitionName, element))
                slot = self.schema.slots[cast(SlotDefinitionName, element)]
                touches.slotrefs.update(set(slot.mixins))
                if slot.is_a:
                    touches.slotrefs.add(slot.is_a)
                if element in self.synopsis.inverses:
                    touches.slotrefs.update(self.synopsis.inverses[cast(SlotDefinitionName, element)])
                if slot.domain:
                    touches.classrefs.add(slot.domain)
                if slot.range in self.schema.classes:
                    touches.classrefs.add(cast(ClassDefinitionName, slot.range))
                elif slot.range in self.schema.types:
                    touches.typerefs.add(cast(TypeDefinitionName, slot.range))
                if slot.in_subset:
                    touches.subsetrefs.update(slot.in_subset)
                for sv in self.schema.slots.values():
                    if sv.is_a == element:
                        touches.slotrefs.add(sv.name)
            if element in self.schema.types:
                touches.typerefs.add(cast(TypeDefinitionName, element))
                typ = self.schema.types[cast(TypeDefinitionName, element)]
                if element in self.synopsis.rangerefs:
                    touches.slotrefs.update(self.synopsis.rangerefs[element])
                if typ.typeof:
                    touches.typerefs.add(cast(TypeDefinitionName, typ.typeof))
                if typ.in_subset:
                    touches.subsetrefs.update(typ.in_subset)
                for tv in self.schema.types.values():
                    if tv.typeof == element:
                        touches.slotrefs.add(tv.name)
            if element in self.schema.subsets:
                touches.subsetrefs.add(cast(SubsetDefinitionName, element))
                if element in self.synopsis.subsetrefs:
                    touches.update(self.synopsis.subsetrefs[cast(SubsetDefinitionName, element)])
            if not bool(touches):
                self.logger.warning(f"neighborhood({element}) - {element} is undefined")

        return touches

    def range_type_path(self, typ: TypeDefinition) -> list[str]:
        """
        Return a formatted list of range types from the base up

        :param typ: type definition whose name is to be formatted
        :return: List of possible types with base at the leftmost
        """
        formatted_typ_name = self.class_or_type_name(typ.name)
        if typ.typeof:
            return self.range_type_path(self.schema.types[cast(TypeDefinitionName, typ.typeof)]) + [formatted_typ_name]
        elif typ.repr:
            return [typ.repr, formatted_typ_name]
        else:
            return [formatted_typ_name]

    def class_identifier(
        self, def_or_name: Union[str, ClassDefinition, TypeDefinition]
    ) -> Optional[SlotDefinitionName]:
        """
        Return the class identifier if any

        :param def_or_name: class name or definition
        :return: name of class key (or identifier) if one exists
        """
        if isinstance(def_or_name, ClassDefinition):
            cls = def_or_name
        elif def_or_name in self.schema.classes:
            cls = self.schema.classes[cast(ClassDefinitionName, def_or_name)]
        else:
            return None
        for slotname in cls.slots:
            slot = self.schema.slots[slotname]
            if slot.identifier or slot.key:
                return slotname
        return None

    @staticmethod
    def enum_identifier_path(enum_or_enumname: Union[str, EnumDefinition]) -> list[str]:
        """Return an enum_identifier path"""
        return [
            "str",
            camelcase(enum_or_enumname.name if isinstance(enum_or_enumname, EnumDefinition) else enum_or_enumname),
        ]

    def class_identifier_path(self, cls_or_clsname: Union[str, ClassDefinition], force_non_key: bool) -> list[str]:
        """
        Return the path closure to a class identifier if the class has a key and force_non_key is false otherwise
        return a dictionary closure.

        :param cls_or_clsname: class definition
        :param force_non_key: True means inlined even if the class has a key
        :return: path
        """
        cls = (
            cls_or_clsname
            if isinstance(cls_or_clsname, ClassDefinition)
            else self.schema.classes[ClassDefinitionName(cls_or_clsname)]
        )

        # Determine whether the class has a key
        identifier_slot = None
        if not force_non_key:
            identifier_slot = self.class_identifier(cls)

        # No key or inlined, its closure is a dictionary
        if identifier_slot is None:
            return ["dict", self.class_or_type_name(cls.name)]

        # We're dealing with a reference
        pathname = camelcase(cls.name + " " + self.aliased_slot_name(identifier_slot))
        if cls.is_a:
            parent_identifier_slot = self.class_identifier(cls.is_a)
            if parent_identifier_slot:
                return self.class_identifier_path(cls.is_a, False) + [pathname]
        return self.slot_range_path(identifier_slot) + [pathname]

    def slot_range_path(self, slot_or_name: Union[str, SlotDefinition]) -> list[str]:
        """
        Return an ordered list of slot ranges from distal to proximal

        :param slot_or_name: slot whose range is being typed
        :return: ordered list of types from base type forward
        """
        slot = (
            slot_or_name
            if isinstance(slot_or_name, SlotDefinition)
            else self.schema.slots[cast(SlotDefinitionName, slot_or_name)]
        )
        if slot.range in self.schema.types:
            # Type
            return self.range_type_path(self.schema.types[cast(TypeDefinitionName, slot.range)])
        elif slot.range in self.schema.enums:
            return self.enum_identifier_path(slot.range)
        else:
            # Class
            return self.class_identifier_path(slot.range, bool(slot.inlined))

    def aliased_slot_name(self, slot: Union[SlotDefinitionName, SlotDefinition]) -> SlotDefinitionName:
        """Return the overloaded slot name -- the alias if one exists otherwise the actual name

        @param slot: either a slot name or a definition
        @return: overloaded name
        """
        if isinstance(slot, str):
            slot = self.schema.slots[cast(SlotDefinitionName, slot)]
        return slot.alias if slot.alias else slot.name

    def class_or_type_for(self, name: str) -> Optional[Element]:
        """
        Return the corresponding class or type for name
        """
        if name in self.schema.classes:
            return self.schema.classes[ClassDefinitionName(name)]
        elif name in self.schema.types:
            return self.schema.types[TypeDefinitionName(name)]
        elif name in self.schema.enums:
            return self.schema.enums[EnumDefinitionName(name)]
        return None

    def class_or_type_name(self, name: str) -> str:
        """
        Return the camelcase representation of clsname if it is a valid class or type.  Prepend "Unknown"
        if the name isn't valid
        """
        if name in self.schema.classes:
            return camelcase(name)
        elif name in self.schema.types:
            typ = self.schema.types[cast(TypeDefinitionName, name)]
            if typ.typeof:
                return camelcase(name)
            else:
                return typ.base
        else:
            return "Unknown_" + camelcase(name)

    def slot_for(self, name: str) -> Optional[Element]:
        return self.schema.slots.get(name)

    def slot_name(self, name: str) -> str:
        """
        Return the underscored version of the aliased slot name if name is a slot. Prepend ``unknown_`` if the name
        isn't valid.
        """
        slot = self.slot_for(name)
        return underscore(self.aliased_slot_name(slot) if slot else ("unknown " + name))

    def subset_for(self, name: str) -> Optional[Element]:
        return self.schema.subsets.get(name)

    def subset_name(self, name: str) -> str:
        subset = self.subset_for(name)
        return ("" if subset else "Unknown_") + camelcase(name)

    def formatted_element_name(
        self, el_or_elname: Union[ElementName, Element], is_range_name: bool = False
    ) -> Optional[str]:
        """
        Return the default format for the name or the referenced element.  Slots are under_scored, all others are
        CamelCased. Slot names are the alias, not the actual name

        :param el_or_elname: element or name to map
        :param is_range_name: True means that we're looking for a class or type.  False means Slot or Subset. Only
            applies if el_or_elname is an ElementName (otherwise we know what we've got
        :return: Formatted name if type can be known else None
        """
        if isinstance(el_or_elname, str):
            if is_range_name:
                return (
                    self.class_or_type_name(el_or_elname)
                    if el_or_elname in self.schema.classes
                    or el_or_elname in self.schema.types
                    or el_or_elname == self.schema.default_range
                    else None
                )
            elif el_or_elname in self.schema.slots:
                return self.slot_name(cast(SlotDefinitionName, el_or_elname))
            elif el_or_elname in self.schema.subsets:
                return self.subset_name(el_or_elname)
            return None
        elif isinstance(el_or_elname, (ClassDefinition, TypeDefinition)):
            return self.class_or_type_name(el_or_elname.name)
        elif isinstance(el_or_elname, SlotDefinition):
            return self.slot_name(el_or_elname.name)
        elif isinstance(el_or_elname, SubsetDefinition):
            return self.subset_name(el_or_elname.name)
        else:
            return None

    def obj_for(self, el_or_elname: str, is_range_name: bool = False) -> Optional[Element]:
        if is_range_name:
            return (
                self.class_or_type_for(el_or_elname)
                if el_or_elname in self.schema.classes
                or el_or_elname in self.schema.types
                or el_or_elname == self.schema.default_range
                else None
            )
        elif el_or_elname in self.schema.slots:
            return self.slot_for(cast(SlotDefinitionName, el_or_elname))
        elif el_or_elname in self.schema.subsets:
            return self.subset_for(el_or_elname)
        else:
            return self.class_or_type_for(el_or_elname)

    def default_prefix(self) -> Optional[str]:
        """Return the default prefix for the schema

        @return: URI or NCNAME of default prefix"""
        if "://" in self.schema.default_prefix:
            return self.schema.default_prefix
        else:
            # Basic loader tests for valid default prefix
            return self.schema.prefixes[PrefixPrefixPrefix(self.schema.default_prefix)].prefix_reference

    # TODO: add lru cache once we get identity into the classes
    def domain_slots(self, cls: ClassDefinition) -> list[SlotDefinition]:
        """Return all slots in the class definition that are owned by the class"""
        domain_slots = []
        for slot_name in cls.slots:
            slot = self.schema.slots[slot_name]

            # add any mixin ancestors here so that slots will be distributed to descendents correctly via mixin
            # hierarchy.
            mixin_ancestors = []
            if cls.mixins:
                for mixin in cls.mixins:
                    for ancestor in self.schemaview.class_ancestors(mixin, mixins=False):
                        if ancestor not in mixin_ancestors:
                            mixin_ancestors.append(ancestor)

            for mixin_ancestor in mixin_ancestors:
                if mixin_ancestor not in cls.mixins:
                    cls.mixins.append(mixin_ancestor)

            # Check if the class is in the domain of the slot or if any of its mixins are in the domain
            if cls.name in slot.domain_of or (set(cls.mixins).intersection(slot.domain_of)):
                domain_slots.append(slot)

        return domain_slots

    def add_mappings(self, defn: Definition) -> None:
        """
        Process any mappings in defn, adding all of the mappings prefixes to the namespace map
        :param defn: Class or Slot Definition
        """
        self.add_id_prefixes(defn)
        mappings = (
            defn.mappings
            + defn.related_mappings
            + defn.close_mappings
            + defn.narrow_mappings
            + defn.broad_mappings
            + defn.exact_mappings
        )
        # see https://github.com/linkml/linkml/pull/283
        if isinstance(defn, ClassDefinition):
            mappings.append(defn.class_uri)
        if isinstance(defn, SlotDefinition):
            mappings.append(defn.slot_uri)
        for mapping in mappings:
            if "://" in str(mapping):
                mcurie = self.namespaces.curie_for(mapping)
                if mcurie is None:
                    self.logger.warning(f"No namespace defined for URI: {mapping}")
                    return  # Absolute path - no prefix/name
                else:
                    mapping = mcurie
            if ":" not in mapping or len(mapping.split(":")) != 2:
                raise ValueError(f"Definition {defn.name} - unrecognized mapping: {mapping}")
            ns = mapping.split(":")[0]
            self.logger.debug(f"Adding {ns} from {mapping} // {defn}")
            if ns:
                self.add_prefix(ns)

    def add_id_prefixes(self, element: Element) -> None:
        for id_prefix in element.id_prefixes:
            self.add_prefix(id_prefix)

    def add_prefix(self, ncname: str) -> None:
        """Add a prefix to the list of prefixes to emit

        @param ncname: name to add
        """
        if ncname not in self.namespaces:
            self.logger.warning(f"Unrecognized prefix: {ncname}")
            self.namespaces[ncname] = f"http://example.org/UNKNOWN/{ncname}/"
        self.emit_prefixes.add(ncname)

    def get_metamodel_slot_name(self, slot_name: str) -> str:
        """
        Allows for localization of some generators, such as markdown generator

        The client can control how metamodel elements are displayed; e.g.
        mapping the metamodel slot "slot" to "field"

        This method also takes care of pluralization; e.g. if "slot" is mapped to
        "field", then "slots" will be mapped to "fields"

        :param slot_name:
        :return:
        """
        is_capitalized = slot_name[0].isupper()

        def capitalize(s: str):
            if is_capitalized:
                return f"{s[0].upper()}{s[1:]}"
            else:
                return s

        slot_name_normalized = underscore(slot_name).lower()
        slot_name_normalized_singular = re.sub(r"s$", "", slot_name_normalized)
        if slot_name_normalized == "classes":
            slot_name_normalized_singular = "class"
        if self.metamodel_name_map is not None and slot_name_normalized in self.metamodel_name_map:
            return capitalize(self.metamodel_name_map[slot_name_normalized])
        elif self.metamodel_name_map is not None and slot_name_normalized_singular in self.metamodel_name_map:
            return capitalize(f"{self.metamodel_name_map[slot_name_normalized_singular]}s")
        else:
            return slot_name

    @staticmethod
    def is_class_unconstrained(cls: ClassDefinition):
        """
        Determine if the class is mapped to typing.Any, i.e the unconstrained class

        :param cls: class definition
        :return: true if the class is unconstrained
        """
        return cls.class_uri == "linkml:Any"


def shared_arguments(g: type[Generator]) -> Callable[[Command], Command]:
    def verbosity_callback(ctx, param, verbose):
        if verbose >= 2:
            logging.basicConfig(level=logging.DEBUG, force=True)
        elif verbose == 1:
            logging.basicConfig(level=logging.INFO, force=True)

    def stacktrace_callback(ctx, param, stacktrace):
        if not stacktrace:
            sys.tracebacklimit = 0

    def decorator(f: Command) -> Command:
        f.params.append(Argument(("yamlfile",), type=click.Path(exists=True, dir_okay=False)))
        f.params.append(
            Option(
                ("--format", "-f"),
                type=click.Choice(g.valid_formats),
                default=g.valid_formats[0],
                show_default=True,
                help="Output format",
            )
        )
        f.params.append(
            Option(
                ("--metadata/--no-metadata",),
                default=True,
                show_default=True,
                help="Include metadata in output",
            )
        )
        f.params.append(
            Option(
                ("--useuris/--metauris",),
                default=True,
                show_default=True,
                help="Use class and slot URIs over model uris",
            )
        )
        f.params.append(Option(("--importmap", "-im"), type=click.File(), help="Import mapping file"))
        log_level_option(f)
        f.params.append(
            Option(
                ("--verbose", "-v"),
                count=True,
                help="Verbosity. Takes precedence over --log_level.",
                callback=verbosity_callback,
            )
        )
        f.params.append(
            Option(
                ("--mergeimports/--no-mergeimports",),
                default=True,
                help="Merge imports into source file (default=mergeimports)",
            )
        )
        f.params.append(
            Option(
                ("--stacktrace/--no-stacktrace",),
                default=False,
                show_default=True,
                help="Print a stack trace when an error occurs",
                callback=stacktrace_callback,
            )
        )

        return f

    return decorator



from linkml_runtime.linkml_model import (
    ClassDefinition,
    EnumDefinitionName,
    SlotDefinition,
)

from linkml.generators.common.ifabsent_processor import IfAbsentProcessor


class PythonIfAbsentProcessor(IfAbsentProcessor):
    UNIMPLEMENTED_DEFAULT_VALUES = ["class_curie", "class_uri", "slot_uri", "slot_curie", "default_range", "default_ns"]

    def map_custom_default_values(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition) -> (bool, str):
        if default_value in self.UNIMPLEMENTED_DEFAULT_VALUES:
            return True, None

        if default_value == "bnode":
            return True, "bnode()"

        return False, None

    def map_string_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return self._strval(default_value)

    def map_boolean_true_default_value(self, slot: SlotDefinition, cls: ClassDefinition):
        return "True"

    def map_boolean_false_default_value(self, slot: SlotDefinition, cls: ClassDefinition):
        return "False"

    def map_integer_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return default_value

    def map_float_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return default_value

    def map_double_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return default_value

    def map_decimal_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return default_value

    def map_time_default_value(self, hour: str, minutes: str, seconds: str, slot: SlotDefinition, cls: ClassDefinition):
        return f"time({int(hour)}, {int(minutes)}, {int(seconds)})"

    def map_date_default_value(self, year: str, month: str, day: str, slot: SlotDefinition, cls: ClassDefinition):
        return f"date({int(year)}, {int(month)}, {int(day)})"

    def map_datetime_default_value(
        self,
        year: str,
        month: str,
        day: str,
        hour: str,
        minutes: str,
        seconds: str,
        slot: SlotDefinition,
        cls: ClassDefinition,
    ):
        return f"datetime({int(year)}, {int(month)}, {int(day)}, {int(hour)}, {int(minutes)}, {int(seconds)})"

    def map_uri_or_curie_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return self._uri_for(default_value)

    def map_curie_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return self._uri_for(default_value)

    def map_uri_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        return self._uri_for(default_value)

    def map_enum_default_value(
        self, enum_name: EnumDefinitionName, permissible_value_name: str, slot: SlotDefinition, cls: ClassDefinition
    ):
        return f"'{permissible_value_name}'"

    def map_nc_name_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_object_identifier_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_node_identifier_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_json_pointer_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_json_path_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    def map_sparql_path_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()



import csv
import inspect
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Union

import click
import linkml_runtime.linkml_model.meta as metamodel
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.introspection import package_schemaview
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.ext.associationproxy import _AssociationCollection
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import StaticPool

from linkml._version import __version__
from linkml.generators.pythongen import PythonGenerator
from linkml.generators.sqlalchemygen import SQLAlchemyGenerator, TemplateEnum
from linkml.generators.sqltablegen import SQLTableGenerator
from linkml.utils import datautils, validation
from linkml.utils.datautils import (
    _get_context,
    _get_format,
    _is_xsv,
    dumpers_loaders,
    get_dumper,
    get_loader,
    infer_index_slot,
    infer_root_class,
)

logger = logging.getLogger(__name__)


@dataclass
class SQLStore:
    """
    A wrapper for a SQLite database.

    This provides two core operations for storing and retrieving data

    - :meth:`dump`
    - :meth:`load`

    The wrapper transparently will take care of:

    - mapping your LinkML schema into SQL Tables
    - creating a SQL Alchemy ORM layer
    - mapping your data/objects in any LinkML compliant data format (json. yaml, rdf) into ORM objects
    """

    schema: Optional[Union[str, Path, SchemaDefinition]] = None
    schemaview: Optional[SchemaView] = None
    engine: Optional[Engine] = None
    database_path: Union[str, Path] = None
    use_memory: bool = False
    """https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#using-a-memory-database-in-multiple-threads"""

    module: Optional[ModuleType] = None
    native_module: Optional[ModuleType] = None
    include_schema_in_database: bool = False

    def __post_init__(self):
        if self.database_path is None and not self.use_memory:
            raise ValueError("Must have database path or use_memory must be True")
        if self.schema is not None and self.schemaview is None:
            self.schemaview = SchemaView(self.schema)

    def db_exists(self, create=True, force=False) -> Optional[str]:
        """
        check if database exists, optionally create if not present

        :param create: create if does not exist
        :param force: recreate database, destroying any content if previously present
        :return: path
        """
        if self.use_memory:
            db_exists = False
        else:
            if not self.database_path:
                raise ValueError("database_path not set")
            db_exists = os.path.exists(self.database_path)
        if force or (create and not db_exists):
            if self.use_memory:
                self.engine = create_engine(
                    "sqlite://", connect_args={"check_same_thread": False}, poolclass=StaticPool
                )
            else:
                if force:
                    Path(self.database_path).unlink(missing_ok=True)
                self.engine = create_engine(f"sqlite:///{self.database_path}")
            with self.engine.connect() as con:
                ddl = SQLTableGenerator(self.schema).generate_ddl()
                con.connection.executescript(ddl)
                if self.include_schema_in_database:
                    metamodel_sv = package_schemaview(metamodel.__name__)
                    meta_ddl = SQLTableGenerator(metamodel_sv.schema).generate_ddl()
                    con.connection.executescript(meta_ddl)
        if self.use_memory:
            return None
        if not os.path.exists(self.database_path):
            raise ValueError(f"No database: {self.database_path}")
        return self.database_path

    def compile(self) -> ModuleType:
        """
        Compile SQLAlchemy object model

        Uses Declarative by default.

        Note that the SQLA model is different from the native dataclass model

        :return: compiled module
        """
        gen = SQLAlchemyGenerator(self.schema)
        self.module = gen.compile_sqla(template=TemplateEnum.DECLARATIVE)
        return self.module

    def compile_native(self) -> ModuleType:
        """
        Compile native python object model

        :return: compiled module
        """
        gen = PythonGenerator(yaml_dumper.dumps(self.schema))
        self.native_module = gen.compile_module()
        return self.native_module

    def load(self, target_class: Union[str, type[YAMLRoot]] = None) -> YAMLRoot:
        """
        Loads a LinkML object from the wrapped SQLite database

        :param target_class:
        :return:
        """
        return self.load_all(target_class=target_class)[0]

    def load_all(self, target_class: Union[str, type[YAMLRoot]] = None) -> list[YAMLRoot]:
        if target_class is None:
            target_class_name = infer_root_class(self.schemaview)
            target_class = self.native_module.__dict__[target_class_name]
        if self.engine is None:
            self.engine = create_engine(f"sqlite:///{self.database_path}")
        session_class = sessionmaker(bind=self.engine)
        with session_class.begin() as session:
            typ = self.to_sqla_type(target_class)
            q = session.query(typ)
            all_objs = q.all()
            tmp = self.from_sqla(all_objs)
        return tmp

    def dump(self, element: YAMLRoot, append=True) -> None:
        """
        Store an element in the database

        :param element:
        :param append:
        :return:
        """
        if self.engine is None:
            raise ValueError("Must set self.engine")
        session_class = sessionmaker(bind=self.engine)
        with session_class.begin() as session:
            nu_obj = self.to_sqla(element)
            if not append:
                session.query(type(nu_obj)).delete()
            session.add(nu_obj)
            session.commit()

    def to_sqla_type(self, target_class: type[YAMLRoot]) -> Any:
        for n, nu_typ in inspect.getmembers(self.module):
            if n == target_class.__name__:
                return nu_typ
        raise ValueError(f"Could not find: {target_class}")

    def from_sqla_type(self, typ) -> Any:
        for n, nu_typ in inspect.getmembers(self.native_module):
            if n == typ.__name__:
                return nu_typ
        raise ValueError(f"Could not find: {typ}")

    def to_sqla(self, obj: Union[YAMLRoot, list]) -> Any:
        """
        Translate native LinkML object to SQLAlchemy declarative module

        :param obj:
        :return:
        """
        if self.module is None:
            self.compile()
        if isinstance(obj, list):
            nu_obj = [self.to_sqla(x) for x in obj]
            if nu_obj:
                return nu_obj
            else:
                return None
        elif isinstance(obj, dict):
            nu_obj = {}
            for k, v in obj.items():
                v2 = self.to_sqla(v)
                if v2 is not None:
                    nu_obj[k] = v2
            if nu_obj:
                return nu_obj
            else:
                return None
        # elif isinstance(obj, PermissibleValue):
        #    return str(obj.text)
        elif isinstance(obj, EnumDefinitionImpl):
            return str(obj)
        elif isinstance(obj, YAMLRoot) or isinstance(obj, BaseModel):
            typ = type(obj)
            inst_args = {}
            for k, v in vars(obj).items():
                v2 = self.to_sqla(v)
                if v2 is not None:
                    inst_args[k] = v2
            for n, nu_typ in inspect.getmembers(self.module):
                # TODO: make more efficient
                if n == typ.__name__:
                    nu_obj = nu_typ(**inst_args)
                    return nu_obj
            raise ValueError(f"Cannot find {typ.__name__} in {self.module}")
        else:
            return obj

    def from_sqla(self, obj: Any) -> Optional[Union[YAMLRoot, list[YAMLRoot]]]:
        """
        Translate from SQLAlchemy declarative module to native LinkML

        :param obj: sqla object
        :return: native dataclass object
        """
        typ = type(obj)
        nm = self.schemaview.class_name_mappings()
        if typ.__name__ in nm:
            cls = nm[typ.__name__]
        else:
            cls = None
        if isinstance(obj, list) or isinstance(obj, _AssociationCollection):
            nu_obj = [self.from_sqla(x) for x in obj]
            if nu_obj:
                return nu_obj
            else:
                return None
        elif cls:
            inst_args = {}
            for sn in self.schemaview.class_slots(cls.name):
                sn = underscore(sn)
                v = getattr(obj, sn, None)
                v2 = self.from_sqla(v)
                if v2 is not None and v2 != [] and v2 != {}:
                    inst_args[sn] = v2
            for n, nu_typ in inspect.getmembers(self.native_module):
                # TODO: make more efficient
                if n == typ.__name__:
                    nu_obj = nu_typ(**inst_args)
                    return nu_obj
            raise ValueError(f"Cannot find {typ.__name__} in {self.native_module}")
        else:
            return obj


@click.group(name="sqldb")
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.option(
    "--csv-field-size-limit",
    type=int,
    help="""
              Increase the default limit for maximum field size.
              See https://docs.python.org/3/library/csv.html#csv.field_size_limit""",
)
@click.version_option(__version__, "-V", "--version")
def main(verbose: int, quiet: bool, csv_field_size_limit: int):
    """Run the LinkML SQL CLI."""
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if quiet:
        logging.basicConfig(level=logging.ERROR)
    if csv_field_size_limit:
        csv.field_size_limit(csv_field_size_limit)


@main.command()
@click.option("--module", "-m", help="Path to python datamodel module")
@click.option("--db", "-D", help="Path to SQLite database file")
@click.option(
    "--input-format",
    "-f",
    type=click.Choice(list(dumpers_loaders.keys())),
    help="Input format. Inferred from input suffix if not specified",
)
@click.option(
    "--target-class",
    "-C",
    help="name of class in datamodel that the root node instantiates",
)
@click.option("--index-slot", "-S", help="top level slot. Required for CSV dumping/loading")
@click.option("--schema", "-s", help="Path to schema specified as LinkML yaml")
@click.option(
    "--validate/--no-validate",
    default=True,
    show_default=True,
    help="Validate against the schema",
)
@click.option(
    "--force/--no-force",
    default=True,
    show_default=True,
    help="Force creation of a database if it does not exist",
)
@click.option(
    "--glob/--no-glob",
    default=False,
    show_default=True,
    help="Treat input as a quoted glob expression, e.g. 'data/*.json'",
)
@click.argument("inputs", nargs=-1)
def dump(
    inputs,
    module,
    db,
    target_class,
    input_format=None,
    schema=None,
    validate=None,
    force: bool = None,
    glob: bool = None,
    index_slot=None,
) -> None:
    """
    Dumps data to a SQL store

    Examples:

        linkml-sqldb dump -s personinfo.yaml -D my_person_database.db personinfo_data01.yaml
    """
    if module is None:
        if schema is None:
            raise Exception("must pass one of module OR schema")
        else:
            python_module = PythonGenerator(schema).compile_module()
    else:
        python_module = compile_python(module)
    if schema is not None:
        sv = SchemaView(schema)
    if target_class is None:
        if sv is None:
            raise ValueError("Must specify schema if not target class is specified")
        target_class = infer_root_class(sv)
    if target_class is None:
        raise Exception("target class not specified and could not be inferred")
    py_target_class = python_module.__dict__[target_class]

    endpoint = SQLStore(schema, database_path=db, include_schema_in_database=False)
    endpoint.native_module = python_module
    endpoint.db_exists(force=force)
    endpoint.compile()

    if glob:
        import glob

        inputs = [item for input in inputs for item in glob.glob(input)]
    for input in inputs:
        logger.info(f"Loading: {input}")
        input_format = _get_format(input, input_format)
        loader = get_loader(input_format)

        inargs = {}
        if datautils._is_rdf_format(input_format):
            if sv is None:
                raise Exception("Must pass schema arg")
            inargs["schemaview"] = sv
            inargs["fmt"] = input_format
        if _is_xsv(input_format):
            if index_slot is None:
                index_slot = infer_index_slot(sv, target_class)
                if index_slot is None:
                    raise Exception("--index-slot is required for CSV input")
            inargs["index_slot"] = index_slot
            inargs["schema"] = schema
        obj = loader.load(source=input, target_class=py_target_class, **inargs)
        if validate:
            if schema is None:
                raise Exception("--schema must be passed in order to validate. Suppress with --no-validate")
            # TODO: use validator framework
            validation.validate_object(obj, schema)
        endpoint.dump(obj)


@main.command()
@click.option("--module", "-m", help="Path to python datamodel module")
@click.option("--db", "-D", help="Path to SQLite database file")
@click.option("--output", "-o", help="Path to output file")
@click.option(
    "--output-format",
    "-t",
    type=click.Choice(list(dumpers_loaders.keys())),
    help="Output format. Inferred from output suffix if not specified",
)
@click.option(
    "--target-class",
    "-C",
    help="name of class in datamodel that the root node instantiates",
)
@click.option("--index-slot", "-S", help="top level slot. Required for CSV dumping/loading")
@click.option("--schema", "-s", help="Path to schema specified as LinkML yaml")
@click.option(
    "--validate/--no-validate",
    default=True,
    show_default=True,
    help="Validate against the schema",
)
@click.option(
    "--force/--no-force",
    default=True,
    show_default=True,
    help="Force creation of a database if it does not exist",
)
def load(
    output,
    output_format,
    module,
    db,
    target_class,
    input_format=None,
    schema=None,
    validate=None,
    force: bool = None,
    index_slot=None,
) -> None:
    """
    Loads data from a SQL store

    Examples:

        linkml-sqldb load -s personinfo.yaml -D my_person_database.db -o my_data.yaml
    """
    if module is None:
        if schema is None:
            raise Exception("must pass one of module OR schema")
        else:
            python_module = PythonGenerator(schema).compile_module()
    else:
        python_module = compile_python(module)
    if schema is not None:
        sv = SchemaView(schema)
    if target_class is not None:
        py_target_class = python_module.__dict__[target_class]
    else:
        py_target_class = None

    endpoint = SQLStore(schema, database_path=db, include_schema_in_database=False)
    endpoint.native_module = python_module
    endpoint.compile()
    obj = endpoint.load(py_target_class)
    output_format = _get_format(output, output_format, default="json")
    outargs = {}
    if output_format == "json-ld":
        if schema is not None:
            context = [_get_context(schema)]
        else:
            raise Exception("Must pass in context OR schema for RDF output")
        if context:
            outargs["contexts"] = list(context)
        outargs["fmt"] = "json-ld"
    if output_format == "rdf" or output_format == "ttl":
        if sv is None:
            raise Exception("Must pass schema arg")
        outargs["schemaview"] = sv
    if _is_xsv(output_format):
        if index_slot is None:
            index_slot = infer_index_slot(sv, target_class)
            if index_slot is None:
                raise Exception("--index-slot is required for CSV output")
        outargs["index_slot"] = index_slot
        outargs["schema"] = schema
    dumper = get_dumper(output_format)
    if output is not None:
        dumper.dump(obj, output, **outargs)
    else:
        print(dumper.dumps(obj, **outargs))


if __name__ == "__main__":
    main()



import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import click
from jinja2 import Template
from linkml_runtime.linkml_model.meta import Prefix
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)

template = """
{% for pfxn, pfx in schema.prefixes.items() -%}
PREFIX {{pfxn}}: <{{pfx.prefix_reference}}>
{% endfor %}

{% for cn, c in schema.classes.items() if not c.mixin and not c.abstract %}

## --
## Checks for {{ cn }}
## --

# @CHECK permitted_{{cn}}
SELECT ?g ?s ?p WHERE {
 GRAPH ?g {
  ?s rdf:type {{ schema_view.get_uri(cn) }} ;
     ?p ?o .
  FILTER ( ?p NOT IN (
   {% for sn in schema_view.class_slots(cn) -%}
   {{ schema_view.get_uri(schema_view.get_slot(sn, attributes=True)) }},
   {% endfor -%}
   rdf:type ))
 }
 {{ extra }}
} {{ limit }}


{% for slot in schema_view.class_induced_slots(cn) -%}

{% if slot.required %}
# @CHECK required_{{cn}}_{{slot.name}}
SELECT
  ?check
  ?graph
  ?subject
  ?predicate WHERE {
 GRAPH ?graph {
  ?subject rdf:type {{ schema_view.get_uri(cn) }} .
  FILTER NOT EXISTS { ?subject  {{ schema_view.get_uri(slot) }} ?o  }
 }
 VALUES ?check { linkml:required }
 VALUES ?predicate { {{schema_view.get_uri(slot)}} }
 {{ extra }}
}  {{ limit }}
{% endif %}

{% if slot.range in schema_view.all_classes() %}
# @CHECK object_range_{{cn}}_{{slot.name}}
SELECT
  ?check
  ?graph
  ?subject
  ?predicate
  ?object
WHERE {
 GRAPH ?graph {
  ?subject rdf:type {{ schema_view.get_uri(cn) }} ;
     ?predicate ?object .
  FILTER NOT EXISTS {
    ?object rdf:type ?otype .
    FILTER ( ?otype IN (
      {% for a in schema_view.class_descendants(slot.range) -%}
      {{ schema_view.get_uri(a) }}
      {{ ", " if not loop.last else "" }}
      {% endfor -%} ))
  }
 }
 VALUES ?check { linkml:range }
 VALUES ?predicate { {{ schema_view.get_uri(slot) }}  }
 {{ extra }}
}  {{ limit }}
{% endif %}

{%- endfor %}


## -- End of checks for {{ cn }}
{% endfor %}
"""

x = """
{% for sn in schema_view.class_slots(c.name) %}
     {{ schema.slots[sn].slot_uri }}
   {% endfor %}
"""


def materialize_schema(schemaview: SchemaView):
    schema = schemaview.schema
    if "rdf" not in schema.prefixes:
        schema.prefixes["rdf"] = Prefix("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
    for scn in schemaview.imports_closure():
        for pfxn, pfx in schemaview.schema_map[scn].prefixes.items():
            if pfxn not in schema:
                schema.prefixes[pfxn] = pfx
    for cn, c in schemaview.all_classes().items():
        for a in list(c.attributes.values()):
            schema.slots[a.name] = a
            c.slots.append(a.name)
            del c.attributes[a.name]
    schemaview.set_modified()
    for cn, c in schemaview.all_classes().items():
        for s in schemaview.class_induced_slots(cn):
            if s.name not in c.slots:
                c.slots.append(s.name)
            c.slot_usage[s.name] = s
            s.slot_uri = schemaview.get_uri(s)


@dataclass
class SparqlGenerator(Generator):
    """
    Generates SPARQL queries that can be used for delayed validation
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    valid_formats = ["sparql"]
    visit_all_class_slots = False
    uses_schemaloader = False

    # ObjectVars
    named_graphs: Optional[list[str]] = None
    limit: Optional[int] = None
    sparql: Optional[str] = None

    def __post_init__(self):
        self.schemaview = SchemaView(self.schema)
        materialize_schema(self.schemaview)
        super().__post_init__()
        self.queries = self.generate_sparql(named_graphs=self.named_graphs, limit=self.limit)

    def generate_sparql(self, named_graphs=None, limit: int = None):
        template_obj = Template(template)
        extra = ""
        if named_graphs is not None:
            extra += f"FILTER( ?graph in ( {','.join(named_graphs)} ))"
        logger.info(f"Named Graphs = {named_graphs} // extra={extra}")
        if limit is not None and isinstance(limit, int):
            limit = f"LIMIT {limit}"
        else:
            limit = ""
        sparql = template_obj.render(schema_view=self.schemaview, schema=self.schema, limit=limit, extra=extra)
        self.sparql = sparql
        queries = self.split_sparql(sparql)
        return queries

    def serialize(self, directory=None) -> str:
        if directory is not None:
            Path(directory).mkdir(parents=True, exist_ok=True)
            for qn, q in self.queries.items():
                qpath = os.path.join(directory, f"{qn}.rq")
                with open(qpath, "w", encoding="UTF-8") as stream:
                    stream.write(q)
        return self.sparql

    @staticmethod
    def split_sparql(sparql: str) -> dict[str, str]:
        lines = sparql.split("\n")
        prolog = ""
        queries = defaultdict(str)
        q = None
        for line in lines:
            if line.startswith("# @"):
                q = underscore(line.replace("# @", ""))
                queries[q] = prolog + "\n"
            elif q is None:
                if line.lower().startswith("prefix"):
                    prolog += line + "\n"
            else:
                queries[q] += line + "\n"
        return queries


@shared_arguments(SparqlGenerator)
@click.command(name="sparql")
@click.option("--dir", "-d", help="Directory in which queries will be deposited")
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, dir, **kwargs):
    """Generate SPARQL queries for validation

    This will generate a directory of queries that can be used for QC over a triplestore that
    is conformant to the same LinkML schema.

    Each query in the directory will be of the form

        CHECK_<ConstraintType>_<SchemaElement>.rq

    Example:

        gen-sparql -d ./sparql/ personinfo.yaml
    """
    SparqlGenerator(yamlfile, **kwargs).serialize(directory=dir)


if __name__ == "__main__":
    cli()



"""
Lifestyle methods mixin
"""

from collections.abc import Iterable
from typing import TypeVar

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    EnumDefinition,
    SchemaDefinition,
    SlotDefinition,
    TypeDefinition,
)

from linkml.generators.common.build import ClassResult, EnumResult, RangeResult, SchemaResult, SlotResult, TypeResult
from linkml.generators.common.template import TemplateModel

TSchema = TypeVar("TSchema", bound=SchemaResult)
TClass = TypeVar("TClass", bound=ClassResult)
TSlot = TypeVar("TSlot", bound=SlotResult)
TRange = TypeVar("TRange", bound=RangeResult)
TType = TypeVar("TType", bound=TypeResult)
TEnum = TypeVar("TEnum", bound=EnumResult)
TTemplate = TypeVar("TTemplate", bound=TemplateModel)


class LifecycleMixin:
    """
    Mixin class for implementing lifecycle methods to modify the generation process
    without needing messy overrides and tiptoeing around parent generator classes.

    Not all classes that inherit from the mixin will be able to implement all of the lifecycle methods.
    The generator class should make it clear which methods are supported in its documentation -
    no programmatic checking for lifecycle method support is available on first implementation.

    Generators should **not** define any of the lifecycle methods themselves - the purpose is
    to allow downstream modification of the generation process with clear entrypoints and without
    needing to copy/paste code from the generator/call ``super()`` methods - if the Generator
    is also using the lifecycle method, the point of the mixin is lost.

    The order that these methods are called may vary in different generators, but in general, if
    slots are generated hierarchically within classes as in the pydantic generator...

    * :meth:`.before_generate_schema`
    * :meth:`.before_generate_classes`
    * :meth:`.before_generate_class`
    * :meth:`.before_generate_slots`
    * :meth:`.before_generate_slot`
    * :meth:`.after_generate_slot`
    * :meth:`.after_generate_slots`
    * :meth:`.after_generate_class`
    * :meth:`.after_generate_classes`
    * :meth:`.after_generate_schema`
    * :meth:`.before_render_template`
    * :meth:`.after_render_template`

    (Add other examples of ordering from other generators as implemented)

    Each method should receive a :class:`.SchemaView` to get perspective on the current schema
    being built. The ``before_`` methods should receive and return a single or list of ``Definitions``
    and the ``after_`` methods should receive and return a single or list of the appropriate :class:`.BuildResult`
    type.
    """

    def before_generate_schema(self, schema: SchemaDefinition, sv: SchemaView) -> SchemaDefinition:
        return schema

    def after_generate_schema(self, schema: TSchema, sv: SchemaView) -> TSchema:
        return schema

    def before_generate_class(self, cls: ClassDefinition, sv: SchemaView) -> ClassDefinition:
        return cls

    def after_generate_class(self, cls: TClass, sv: SchemaView) -> TClass:
        return cls

    def before_generate_classes(self, cls: Iterable[ClassDefinition], sv: SchemaView) -> Iterable[ClassDefinition]:
        return cls

    def after_generate_classes(self, cls: Iterable[TClass], sv: SchemaView) -> Iterable[TClass]:
        return cls

    def before_generate_slot(self, slot: SlotDefinition, sv: SchemaView) -> SlotDefinition:
        return slot

    def after_generate_slot(self, slot: TSlot, sv: SchemaView) -> TSlot:
        return slot

    def before_generate_slots(self, slot: Iterable[SlotDefinition], sv: SchemaView) -> Iterable[SlotDefinition]:
        return slot

    def after_generate_slots(self, slot: Iterable[TSlot], sv: SchemaView) -> Iterable[TSlot]:
        return slot

    def before_generate_class_slot(self, slot: SlotDefinition, cls: ClassDefinition, sv: SchemaView) -> SlotDefinition:
        return slot

    def after_generate_class_slot(self, slot: TSlot, cls: ClassDefinition, sv: SchemaView) -> TSlot:
        return slot

    def before_generate_class_slots(
        self, slot: Iterable[SlotDefinition], cls: ClassDefinition, sv: SchemaView
    ) -> Iterable[SlotDefinition]:
        return slot

    def after_generate_class_slots(
        self, slot: Iterable[TSlot], cls: ClassDefinition, sv: SchemaView
    ) -> Iterable[TSlot]:
        return slot

    def before_generate_type(self, typ: TypeDefinition, sv: SchemaView) -> TypeDefinition:
        return typ

    def after_generate_type(self, typ: TType, sv: SchemaView) -> TType:
        return typ

    def before_generate_types(self, typ: Iterable[TypeDefinition], sv: SchemaView) -> Iterable[TypeDefinition]:
        return typ

    def after_generate_types(self, typ: Iterable[TType], sv: SchemaView) -> Iterable[TType]:
        return typ

    def before_generate_enum(self, enum: EnumDefinition, sv: SchemaView) -> EnumDefinition:
        return enum

    def after_generate_enum(self, enum: TEnum, sv: SchemaView) -> TEnum:
        return enum

    def before_generate_enums(self, enum: Iterable[EnumDefinition], sv: SchemaView) -> Iterable[EnumDefinition]:
        return enum

    def after_generate_enums(self, enum: Iterable[TEnum], sv: SchemaView) -> Iterable[TEnum]:
        return enum

    def before_render_template(self, template: TTemplate, sv: SchemaView) -> TTemplate:
        return template

    def after_render_template(self, template: str, sv: SchemaView) -> str:
        return template



class DependencySorter:
    """
    Module for sorting a dependency graph using string labels for nodes.
    This isolates the algorithm from other implementation details.
    The current implementation is intended as a helper for the LinkML Pandera generator
    rather than for general use.
    """

    def __init__(self, dependency_dict: dict[str, list[str]] = None):
        """
        Initialize the DependencySorter with a dictionary representing a dependency graph.

        :param dependency_dict: A dictionary where keys are labels and values are lists of dependency.
        :type dependency_dict: dict[str, list[str]], optional
        """
        self.dependency_dict = dependency_dict or {}

    def _expand_dependencies(self) -> None:
        """
        Explicitly represent nodes with no dependencies as having an empty dependency list.
        This is for easier lookup when visiting nodes.
        """
        nodes_with_dependencies = set(self.dependency_dict)
        all_nodes = set(sum(self.dependency_dict.values(), []))
        no_dependency_nodes = all_nodes - nodes_with_dependencies
        leaf_dependency_graph = {n: [] for n in no_dependency_nodes}
        self.dependency_dict.update(leaf_dependency_graph)

    def _visit(self, node: str, visited: set[str], in_progress: set[str], result: list[str]) -> None:
        """
        recursively visit dependencies in depth first order.
        the graph dicts are assumed to be in proper form (aside from cycles)

        :param visited: tracks progress (initialize to set())
        :param result: builds sorted dependency list (initialize to [])
        :param in_progress: for tracking cycles (initialize to set())
        """
        if node in visited:
            return
        if node in in_progress:
            cycle_path = " -> ".join(list(in_progress) + [node])

            raise ValueError(f"Cyclic dependency detected: {cycle_path}")

        in_progress.add(node)
        for dependency in self.dependency_dict[node]:
            self._visit(dependency, visited, in_progress, result)
        in_progress.remove(node)
        visited.add(node)
        result.append(node)

    def sort_dependencies(self) -> list[str]:
        """
        calculate the order of a dependency graph.

        :return: list of node label strings in dependency order.
        :raises ValueError: If there are cyclic dependencies
        """
        self._expand_dependencies()

        result = []
        visited = set()
        in_progress = set()

        for node in self.dependency_dict:
            if node not in visited:
                self._visit(node, visited, in_progress, result)

        return result

    def add_dependency(self, node: str, dependency: str = None):
        """
        Add a single dependency to the dependency dictionary.
        If dependency is not provided or is None the node is added with an empty
        dependency list.

        :param node: The node label to which the dependency is added.
        :param dependency: The dependency label to be added.
        """
        if node not in self.dependency_dict:
            self.dependency_dict[node] = []

        if dependency is not None and dependency not in self.dependency_dict[node]:
            self.dependency_dict[node].append(dependency)



"""Generate Summary Spreadsheets"""

import os
from csv import DictWriter
from dataclasses import dataclass
from io import StringIO
from typing import Optional

import click
from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class SummaryGenerator(Generator):
    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["tsv"]

    dirname: str = None
    classtab: Optional[DictWriter] = None
    slottab: Optional[DictWriter] = None
    dialect: str = "excel-tab"

    _str_io: Optional[StringIO] = None

    def visit_schema(self, **_) -> None:
        self._str_io = StringIO()
        self.classtab = DictWriter(
            self._str_io,
            [
                "Class Name",
                "Parent Class",
                "YAML Class Name",
                "Description",
                "Flags",
                "Slot Name",
                "YAML Slot Name",
                "Range",
                "Card",
                "Slot Description",
                "URI",
            ],
            dialect=self.dialect,
        )
        self.classtab.writeheader()

    def visit_class(self, cls: ClassDefinition) -> bool:
        self.classtab.writerow(
            {
                "Class Name": camelcase(cls.name),
                "Parent Class": camelcase(cls.is_a) if cls.is_a else "",
                "YAML Class Name": cls.name,
                "Description": cls.description,
            }
        )
        return True

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> None:
        min_card = 1 if slot.required else 0
        max_card = "*" if slot.multivalued else 1
        abstract = "A" if slot.abstract or slot.mixin else ""
        key = "K" if slot.key else ""
        identifier = "I" if slot.identifier else ""
        readonly = "R" if slot.readonly else ""
        ref = "*" if slot.range in self.schema.classes and not slot.inlined else ""
        self.classtab.writerow(
            {
                "Slot Name": aliased_slot_name,
                "Flags": abstract + key + identifier + readonly,
                "Card": f"{min_card}..{max_card}",
                "YAML Slot Name": slot.name if slot.name != aliased_slot_name else "",
                "Range": ref + self.class_or_type_name(slot.range),
                "Slot Description": slot.description,
                "URI": slot.slot_uri,
            }
        )

    def end_schema(self, **kwargs) -> Optional[str]:
        return self._str_io.getvalue()


@shared_arguments(SummaryGenerator)
@click.version_option(__version__, "-V", "--version")
@click.command(name="summary")
def cli(yamlfile, **args):
    """Generate TSV summary files for viewing in Excel and the like"""
    print(SummaryGenerator(yamlfile, **args).serialize(**args))


if __name__ == "__main__":
    cli()



"""Rule implementation for the linkml linter."""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from functools import cache
from itertools import chain

from linkml_runtime.linkml_model import (
    ClassDefinition,
    ClassDefinitionName,
    Element,
    ElementName,
    EnumDefinition,
    EnumDefinitionName,
    SlotDefinition,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.schemaview import SchemaView
from prefixmaps.io.parser import load_multi_context

from linkml import LOCAL_METAMODEL_YAML_FILE
from linkml.linter.config.datamodel.config import (
    CanonicalPrefixesConfig,
    RecommendedRuleConfig,
    RuleConfig,
    StandardNamingConfig,
    TreeRootClassRuleConfig,
)
from linkml.linter.linter import LinterProblem


class LinterRule(ABC):
    PATTERNS = {
        "snake": re.compile(r"[a-z][_a-z0-9]+"),
        "uppersnake": re.compile(r"[A-Z][_A-Z0-9]+"),
        "camel": re.compile(r"[a-z][a-zA-Z0-9]+"),
        "uppercamel": re.compile(r"[A-Z][a-zA-Z0-9]+"),
        "kebab": re.compile(r"[a-z][\-a-z0-9]+"),
        "_uncamel": re.compile(r"(?<!^)(?=[A-Z])"),
    }

    def __init__(self, config: RuleConfig) -> None:
        super().__init__()
        self.config = config

    @property
    @abstractmethod
    def id(self) -> str:
        pass

    @abstractmethod
    def check(self, schema_view: SchemaView, fix: bool) -> Iterable[LinterProblem]:
        pass

    @staticmethod
    def uncamel(n: str) -> str:
        return LinterRule.PATTERNS._uncamel.sub(" ", n)

    @staticmethod
    def format_element(element: Element):
        class_name = element.__class__.__name__.replace("Definition", "")
        return f"{class_name} '{element.name}'"


class NoEmptyTitleRule(LinterRule):
    id = "no_empty_title"

    # todo PVs are not checked for titles yet

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        excluded_types = [t.text if hasattr(t, "text") else str(t) for t in getattr(self.config, "exclude_type", [])]
        for e in schema_view.all_elements(imports=False).values():
            element_type_name = type(e).class_name
            if element_type_name in excluded_types:
                continue
            if fix and e.title is None:
                title = e.name.replace("_", " ")
                title = self.uncamel(title).lower()
                e.title = title
            if e.title is None:
                problem = LinterProblem(message=f"{self.format_element(e)} has no title")
                yield problem


class NoXsdIntTypeRule(LinterRule):
    id = "no_xsd_int_type"

    def check(self, schema_view: SchemaView, fix: bool = False):
        for type_definition in schema_view.all_types(imports=False).values():
            if type_definition.uri == "xsd:int":
                if fix:
                    type_definition.uri = "xsd:integer"
                else:
                    yield LinterProblem(f"{self.format_element(type_definition)} has uri xsd:int")


class PermissibleValuesFormatRule(LinterRule):
    id = "permissible_values_format"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        pattern = self.PATTERNS.get(self.config.format, re.compile(self.config.format))
        for enum_def in schema_view.all_enums(imports=False).values():
            for value in enum_def.permissible_values:
                if pattern.fullmatch(value) is None:
                    yield LinterProblem(f"{self.format_element(enum_def)} has permissible value '{value}'")


@cache
def _get_recommended_metamodel_slots() -> list[str]:
    meta_schema_view = SchemaView(LOCAL_METAMODEL_YAML_FILE)
    recommended_meta_slots = []
    for class_name in meta_schema_view.all_classes(imports=False).keys():
        class_slots = meta_schema_view.class_induced_slots(class_name)
        for slot in class_slots:
            if slot.recommended:
                recommended_meta_slots.append(f"{class_name}__{slot.name}")
    return recommended_meta_slots


class RecommendedRule(LinterRule):
    id = "recommended"

    # todo PVs are not checked for recommended fields yet

    def __init__(self, config: RecommendedRuleConfig) -> None:
        self.config = config

    def check(self, schema_view: SchemaView, fix: bool = False):
        recommended_meta_slots = _get_recommended_metamodel_slots()
        excluded_types = [t.text if hasattr(t, "text") else str(t) for t in getattr(self.config, "exclude_type", [])]
        for element_name, element_definition in schema_view.all_elements(imports=False).items():
            element_type_name = type(element_definition).class_name
            if self.config.include and element_name not in self.config.include:
                continue
            if element_name in self.config.exclude:
                continue
            if element_type_name in excluded_types:
                continue
            for meta_slot_name, meta_slot_value in vars(element_definition).items():
                key = f"{element_definition.class_name}__{meta_slot_name}"
                if key in recommended_meta_slots and not meta_slot_value:
                    yield LinterProblem(
                        f"{self.format_element(element_definition)} does not have recommended slot '{meta_slot_name}'"
                    )


class TreeRootClassRule(LinterRule):
    id = "tree_root_class"

    def __init__(self, config: TreeRootClassRuleConfig) -> None:
        super().__init__(config)

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        tree_roots = [c for c in schema_view.all_classes(imports=False).values() if c.tree_root]
        if len(tree_roots) > 0:
            if self.config.validate_existing_class_name:
                for tree_root in tree_roots:
                    if str(tree_root.name) != self.config.root_class_name:
                        yield LinterProblem(message=f"Tree root class has an invalid name '{tree_root.name}'")
            if len(tree_roots) > 1:
                yield LinterProblem("Schema has more than one class with `tree_root: true`")
        elif fix:
            container = ClassDefinition(self.config.root_class_name, tree_root=True)
            schema_view.add_class(container)
            self.add_index_slots(schema_view, container.name)
        else:
            yield LinterProblem("Schema does not have class with `tree_root: true`")

    def add_index_slots(
        self,
        schema_view: SchemaView,
        container_name: ClassDefinitionName,
        inlined_as_list: bool = False,
        must_have_identifier: bool = False,
        slot_name_func: Callable | None = None,
        convert_camel_case: bool = False,
    ) -> list[SlotDefinition]:
        """Add index slots to a container pointing at all top-level classes.

        :param schema: input schema, will be modified in place
        :param container_name:
        :param inlined_as_list:
        :param must_have_identifier:
        :param slot_name_func: function to determine the name of the slot from the class
        :return: new slots
        """
        container = schema_view.get_class(container_name)
        ranges = {s.range for cn in schema_view.all_classes() for s in schema_view.class_induced_slots(cn)}
        top_level_classes = [c for c in schema_view.all_classes().values() if not c.tree_root and c.name not in ranges]
        if must_have_identifier:
            top_level_classes = [c for c in top_level_classes if schema_view.get_identifier_slot(c.name) is not None]
        index_slots: list[SlotDefinition] = []
        for c in top_level_classes:
            has_identifier = schema_view.get_identifier_slot(c.name)
            if slot_name_func:
                sn = slot_name_func(c)
            else:
                cn = c.name
                if convert_camel_case:
                    cn = self.uncamel(cn).lower()
                cn = cn.replace(" ", "_")
                sn = f"{cn}_index"
            index_slot = SlotDefinition(
                sn,
                range=c.name,
                multivalued=True,
                inlined_as_list=not has_identifier or inlined_as_list,
            )
            index_slots.append(index_slot)
            schema_view.add_slot(index_slot)
            container.slots.append(index_slot.name)
        return index_slots


class NoUndeclaredSlotsRule(LinterRule):
    """Linter rule to check that all slots from a class have been declared in the `slots` section."""

    id = "no_undeclared_slots"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        """Perform the check for undeclared slots.

        :param schema_view: schema to be checked in a SchemaView object.
        :type schema_view: SchemaView
        :param fix: whether or not to fix, defaults to False
        :type fix: bool, optional
        :yield: iterable of error messages about non-compliant slots.
        :rtype: Iterator[Iterable[LinterProblem]]
        """
        all_slots = schema_view.all_slots()
        for class_name in schema_view.all_classes():
            for slot_name in schema_view.class_slots(class_name):
                if slot_name not in all_slots:
                    yield LinterProblem(
                        f"Slot '{slot_name}' from class '{class_name}' not found in schema 'slots' declaration."
                    )


class NoInvalidSlotUsageRule(LinterRule):
    id = "no_invalid_slot_usage"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        for class_name, class_definition in schema_view.all_classes(imports=False).items():
            slot_usage = class_definition.slot_usage
            if not slot_usage:
                continue
            class_slots = schema_view.class_slots(class_name)
            for slot_usage_name in slot_usage:
                if slot_usage_name not in class_slots:
                    yield LinterProblem(f"Slot '{slot_usage_name}' not found on class '{class_name}'")


class NoUndeclaredRangesRule(LinterRule):
    id = "no_undeclared_ranges"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        all_types: dict[TypeDefinitionName, TypeDefinition] = schema_view.all_types()
        all_enums: dict[EnumDefinitionName, EnumDefinition] = schema_view.all_enums()
        all_classes: dict[ClassDefinitionName, ClassDefinition] = schema_view.all_classes()
        all_possible_ranges: set[TypeDefinitionName | EnumDefinitionName | ClassDefinitionName] = (
            set(all_types) | set(all_enums) | set(all_classes)
        )

        # check that the default_range has a valid value
        default_range = schema_view.schema.default_range
        if default_range and default_range not in all_possible_ranges:
            yield LinterProblem(f"Schema default_range '{default_range}' is not defined.")

        for class_name in schema_view.all_classes():
            for slot in schema_view.class_induced_slots(class_name):
                slot_range: set[ElementName] = set(schema_view.slot_range_as_union(slot))

                # check slot range is valid
                for range_name in slot_range:
                    if range_name not in all_possible_ranges:
                        yield LinterProblem(
                            f"Class '{class_name}' slot '{slot.name}' range '{range_name}' is not defined."
                        )


class RootTypeChecks(LinterRule):
    """Performs basic checks on types.

    Types can be defined as subtypes of existing types using the `typeof` attribute.
    Root types are those types that do not inherit from other types using `typeof`.

    - types used in 'typeof' statements must be defined in the types section
    - types cannot be their own 'typeof' parent
    - every root type must have a 'base'
    - every root type must have a 'uri'

    :param LinterRule: linter rule class
    :type LinterRule: LinterRule
    """

    id = "root_type_checks"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        type_ix = schema_view.all_types()
        for t, type_def in type_ix.items():
            if type_def.typeof:
                # This is a child type. Ensure that the parent exists.
                if type_def.typeof not in schema_view.all_types():
                    yield LinterProblem(f"'{t}' has invalid typeof parent '{type_def.typeof}'")
                # Ensure that a type is not its own parent.
                # N.b. at some point, this can/should be extended to checking the full 'typeof' ancestor chain
                # for circular inheritance.
                if type_def.typeof == type_def.name:
                    yield LinterProblem(f"'{t}' has invalid circular typeof parent '{type_def.typeof}'")
                continue
            if not type_def.base:
                yield LinterProblem(f"Root type '{t}' is missing the required 'base' attribute")

            if not type_def.uri:
                yield LinterProblem(f"Root type '{t}' is missing the required 'uri' attribute")


class OnePerClass(LinterRule):
    """Ensures that there is only one slot with the specified attribute per class."""

    id = "one_per_class"

    def _check(self, schema_view: SchemaView, attr_name: str) -> Iterable[LinterProblem]:
        for cn in schema_view.all_classes():
            slots_with_attr = [
                slot
                for slot in schema_view.class_induced_slots(cn)
                if hasattr(slot, attr_name) and getattr(slot, attr_name)
            ]
            if len(slots_with_attr) > 1:
                naughty_slots = ", ".join(sorted([slot.name for slot in slots_with_attr]))
                yield LinterProblem(f"Class '{cn}' has more than one '{attr_name}' slot: {naughty_slots}")


class OneIdentifierPerClass(OnePerClass):
    """Ensure that there is only one slot with the `identifier` attribute set to True per class."""

    id = "one_identifier_per_class"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        """Run the one_identifier_per_class check."""
        return self._check(schema_view, "identifier")


class OneKeyPerClass(OnePerClass):
    """Ensure that there is only one slot with the `key` attribute set to True per class."""

    id = "one_key_per_class"

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        """Run the one_key_per_class check."""
        return self._check(schema_view, "key")


class StandardNamingRule(LinterRule):
    id = "standard_naming"

    def __init__(self, config: StandardNamingConfig) -> None:
        self.config = config

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        excluded_types = [t.text if hasattr(t, "text") else str(t) for t in getattr(self.config, "exclude_type", [])]
        excluded_names = set(getattr(self.config, "exclude", []))

        # element naming patterns indexed by element type
        pattern = {
            "class": (
                self.PATTERNS["uppercamel"]
                if not self.config.class_pattern
                else self.PATTERNS.get(self.config.class_pattern, re.compile(self.config.class_pattern))
            ),
            "slot": (
                self.PATTERNS["snake"]
                if not self.config.slot_pattern
                else self.PATTERNS.get(self.config.slot_pattern, re.compile(self.config.slot_pattern))
            ),
            "enum": self.PATTERNS["uppercamel"],
            "permissible_value": self.PATTERNS["uppersnake"]
            if self.config.permissible_values_upper_case
            else self.PATTERNS["snake"],
        }

        # explicit method mapping for better readability and maintainability
        element_extractors = {
            "class": schema_view.all_classes,
            "slot": schema_view.all_slots,
            "enum": schema_view.all_enums,
        }

        # cache the results to avoid repeated method calls
        cached_elements = {el_type: element_extractors[el_type]() for el_type in element_extractors}

        # generate LinterProblems for classes, slots, and enums incrementally
        problems = (
            LinterProblem(f"{el_type.capitalize()} has name '{el_name}'")
            for el_type, elements in cached_elements.items()
            for el_name in elements
            if f"{el_type}_definition" not in excluded_types
            and el_name not in excluded_names
            and pattern[el_type].fullmatch(el_name) is None
        )

        if "permissible_value" not in excluded_types:
            # add issues from the permissible values, if appropriate
            pv_problems = (
                LinterProblem(f"Permissible value of Enum '{en}' has name '{pv}'")
                for en, en_def in schema_view.all_enums(imports=False).items()
                for pv in en_def.permissible_values
                if pv not in excluded_names and pattern["permissible_value"].fullmatch(pv) is None
            )
            # chain the generators together
            return chain(problems, pv_problems)

        return iter(problems)


class CanonicalPrefixesRule(LinterRule):
    id = "canonical_prefixes"

    def __init__(self, config: CanonicalPrefixesConfig) -> None:
        self.config = config

    def check(self, schema_view: SchemaView, fix: bool = False) -> Iterable[LinterProblem]:
        context = load_multi_context(self.config.prefixmaps_contexts)
        prefix_to_namespace = context.as_dict()
        namespace_to_prefix = context.as_inverted_dict()
        for prefix in schema_view.schema.prefixes.values():
            if (
                prefix.prefix_prefix in prefix_to_namespace
                and prefix.prefix_reference != prefix_to_namespace[prefix.prefix_prefix]
            ):
                yield LinterProblem(
                    f"Schema maps prefix '{prefix.prefix_prefix}' to namespace "
                    f"'{prefix.prefix_reference}' instead of namespace "
                    f"'{prefix_to_namespace[prefix.prefix_prefix]}'"
                )
            if (
                prefix.prefix_reference in namespace_to_prefix
                and prefix.prefix_prefix != namespace_to_prefix[prefix.prefix_reference]
            ):
                yield LinterProblem(
                    f"Schema maps prefix '{prefix.prefix_prefix}' to namespace "
                    f"'{prefix.prefix_reference}' instead of using prefix "
                    f"'{namespace_to_prefix[prefix.prefix_reference]}'"
                )



import inspect
from functools import wraps

import pandera
import polars as pl
from pandera.api.polars.types import PolarsData

from linkml.generators.panderagen.transforms import (
    CollectionDictModelTransform,
    ListDictModelTransform,
    NestedStructModelTransform,
    SimpleDictModelTransform,
)


def handle_validation_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except pl.exceptions.PanicException:
            data = args[2] if len(args) > 2 else kwargs.get("data")
            return data.lazyframe.select(pl.lit(False))
        except pandera.errors.SchemaError as e:
            raise e
        except Exception:
            data = args[2] if len(args) > 2 else kwargs.get("data")
            return data.lazyframe.select(pl.lit(False))

    return wrapper


class LinkmlPanderaValidator:
    @classmethod
    def get_id_column_name(cls):
        return cls._id_name

    @classmethod
    def _simple_dict_fields(cls, column_name):
        details = cls._INLINE_DETAILS[column_name]  # <-- THESE ARE GOING ON THE OUTER CLASS

        return (details["id"], details["other"])

    @classmethod
    def _prepare_simple_dict(cls, data: PolarsData):
        """Returns just the simple dict column transformed to an inlined list form

        note that this method uses collect and iter_rows so is very inefficient
        """
        column_name = data.key
        polars_schema = cls.get_nested_range(column_name).to_schema()
        (id_column, other_column) = cls._simple_dict_fields(column_name)

        simple_dict_transformer = SimpleDictModelTransform(polars_schema, id_column, other_column)

        one_column_df = data.lazyframe.select(pl.col(column_name)).collect()

        list_of_structs = [simple_dict_transformer.transform(e) for [e] in one_column_df.iter_rows()]

        return pl.DataFrame(pl.Series(list_of_structs).alias(column_name))

    @classmethod
    @handle_validation_exceptions
    def _check_simple_dict(cls, data: PolarsData):
        """
        The 'simple dict' format, in which the key serves as a local identifier is not a good match for a PolaRS
        DataFrame. At present the format is
        """
        df = cls._prepare_simple_dict(data)

        column_name = data.key

        polars_schema = cls.get_nested_range(column_name).to_schema()
        simple_transform = SimpleDictModelTransform(polars_schema, *cls._simple_dict_fields(column_name))
        df = simple_transform.explode_unnest_dataframe(df, column_name)

        nested_cls = cls.get_nested_range(column_name)
        nested_cls.validate(df)
        return data.lazyframe.select(pl.lit(True))

    @classmethod
    @handle_validation_exceptions
    def _check_collection_struct(cls, data: PolarsData):
        column_name = data.key
        nested_cls = cls.get_nested_range(column_name)

        df = CollectionDictModelTransform.prepare_dataframe(data, column_name, nested_cls)

        collection_transform = CollectionDictModelTransform(nested_cls.to_schema(), nested_cls.get_id_column_name())
        df = collection_transform.explode_unnest_dataframe(df, column_name)

        nested_cls.validate(df)
        return data.lazyframe.select(pl.lit(True))

    @classmethod
    @handle_validation_exceptions
    def _check_nested_list_struct(cls, data: PolarsData):
        """Use this in a custom check. Pass the nested model as pandera_model."""
        column_name = data.key
        nested_cls = cls.get_nested_range(column_name)

        df = ListDictModelTransform.prepare_dataframe(data, column_name, nested_cls)

        list_transform = ListDictModelTransform(nested_cls.to_schema())
        df = list_transform.explode_unnest_dataframe(df, column_name, data)

        nested_cls.validate(df)
        return data.lazyframe.select(pl.lit(True))

    @classmethod
    @handle_validation_exceptions
    def _check_nested_struct(cls, data: PolarsData):
        """Use this in a custom check. Pass the nested model as pandera_model."""
        column_name = data.key
        nested_cls = cls.get_nested_range(column_name)

        df = NestedStructModelTransform.prepare_dataframe(data, column_name, nested_cls)
        nested_transform = NestedStructModelTransform(nested_cls.to_schema())
        df = nested_transform.explode_unnest_dataframe(df, column_name)

        nested_cls.validate(df)
        return data.lazyframe.select(pl.lit(True))

    @classmethod
    def get_nested_range(cls, column_name):
        """Resolve a nested class range at runtime.

        Nested classes are not stored in the pandera schema,
        but rather in the _NESTED_RANGES dictionary as strings.
        """
        nested_cls_name = cls._NESTED_RANGES[column_name]
        shared_model_module = inspect.getmodule(cls)
        nested_cls = getattr(shared_model_module, nested_cls_name)

        return nested_cls

    @classmethod
    def generate_polars_schema_simple(cls):
        # This is not nesting or list aware, so needs to be aligned with the other method
        return pl.Struct({k: v.dtype.type for k, v in cls.to_schema().columns.items()})

    @classmethod
    def generate_polars_schema(cls, object_to_validate, parser=False) -> dict:
        """Creates a nested PolaRS schema suitable for loading the object_to_validate.
        Optional columns that are not present in the data are omitted.
        This approach is only suitable to enable the test fixtures.
        """
        polars_schema = {}

        if isinstance(object_to_validate, list):
            object_to_validate = object_to_validate[0]

        for column_name, column in cls.to_schema().columns.items():
            dtype = column.properties["dtype"]
            required = column.properties["required"]

            if required or column_name in object_to_validate:
                if dtype.type in [pl.Struct, pl.List]:  # maybe use inline form directly here
                    inline_form = cls._INLINE_FORM.get(column_name, "not_inline")
                    if inline_form == "simple_dict":
                        polars_schema[column_name] = pl.Object  # make this a struct and make the nested non-
                    elif inline_form == "not_inline":
                        polars_schema[column_name] = dtype.type
                    else:
                        nested_cls = cls.get_nested_range(column_name)
                        if inline_form == "inlined_dict":
                            if parser:
                                nested_schema = nested_cls.generate_polars_schema(
                                    object_to_validate[column_name], parser
                                )
                                polars_schema[column_name] = pl.Struct(nested_schema)
                            else:
                                polars_schema[column_name] = pl.Struct
                        elif inline_form == "inlined_list_dict":
                            if parser:
                                nested_schema = nested_cls.generate_polars_schema(
                                    object_to_validate[column_name], parser
                                )
                                polars_schema[column_name] = pl.List(pl.Struct(nested_schema))
                            else:
                                # transformed form
                                polars_schema[column_name] = pl.List
                else:
                    polars_schema[column_name] = dtype.type

        return polars_schema



import inspect
import keyword
import logging
import os
import re
import textwrap
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from types import ModuleType
from typing import ClassVar, Literal, Optional, TypeVar, Union, overload

import click
from jinja2 import ChoiceLoader, Environment, FileSystemLoader, Template
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ElementName,
    SchemaDefinition,
    SlotDefinition,
    TypeDefinition,
)
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.formatutils import camelcase, remove_empty_items, underscore
from linkml_runtime.utils.schemaview import SchemaView
from pydantic.version import VERSION as PYDANTIC_VERSION

from linkml._version import __version__
from linkml.generators.common.lifecycle import LifecycleMixin
from linkml.generators.common.type_designators import get_accepted_type_designator_values, get_type_designator_value
from linkml.generators.oocodegen import OOCodeGenerator
from linkml.generators.pydanticgen import includes
from linkml.generators.pydanticgen.array import ArrayRangeGenerator, ArrayRepresentation
from linkml.generators.pydanticgen.build import ClassResult, SlotResult, SplitResult
from linkml.generators.pydanticgen.pydantic_ifabsent_processor import PydanticIfAbsentProcessor
from linkml.generators.pydanticgen.template import (
    Import,
    Imports,
    ObjectImport,
    PydanticAttribute,
    PydanticBaseModel,
    PydanticClass,
    PydanticModule,
    PydanticTemplateModel,
)
from linkml.utils import deprecation_warning
from linkml.utils.generator import shared_arguments

logger = logging.getLogger(__name__)


if int(PYDANTIC_VERSION[0]) == 1:
    deprecation_warning("pydantic-v1")


def _get_pyrange(t: TypeDefinition, sv: SchemaView) -> str:
    pyrange = t.repr if t is not None else None
    if pyrange is None:
        pyrange = t.base
    if t.base == "XSDDateTime":
        pyrange = "datetime "
    if t.base == "XSDDate":
        pyrange = "date"
    if pyrange is None and t.typeof is not None:
        pyrange = _get_pyrange(sv.get_type(t.typeof), sv)
    if pyrange is None:
        raise Exception(f"No python type for range: {t.name} // {t}")
    return pyrange


DEFAULT_IMPORTS = (
    Imports()
    + Import(module="__future__", objects=[ObjectImport(name="annotations")])
    + Import(
        module="datetime", objects=[ObjectImport(name="datetime"), ObjectImport(name="date"), ObjectImport(name="time")]
    )
    + Import(module="decimal", objects=[ObjectImport(name="Decimal")])
    + Import(module="enum", objects=[ObjectImport(name="Enum")])
    + Import(module="re")
    + Import(module="sys")
    + Import(
        module="typing",
        objects=[
            ObjectImport(name="Any"),
            ObjectImport(name="ClassVar"),
            ObjectImport(name="Literal"),
            ObjectImport(name="Optional"),
            ObjectImport(name="Union"),
        ],
    )
    + Import(
        module="pydantic",
        objects=[
            ObjectImport(name="BaseModel"),
            ObjectImport(name="ConfigDict"),
            ObjectImport(name="Field"),
            ObjectImport(name="RootModel"),
            ObjectImport(name="SerializationInfo"),
            ObjectImport(name="SerializerFunctionWrapHandler"),
            ObjectImport(name="field_validator"),
            ObjectImport(name="model_serializer"),
        ],
    )
)

DEFAULT_INJECTS = [includes.LinkMLMeta]


class MetadataMode(str, Enum):
    FULL = "full"
    """
    all metadata from the source schema will be included, even if it is represented by the template classes,
    and even if it is represented by some child class (eg. "classes" will be included with schema metadata
    """
    EXCEPT_CHILDREN = "except_children"
    """
    all metadata from the source schema will be included, even if it is represented by the template classes,
    except if it is represented by some child template class (eg. "classes" will be excluded from schema metadata)
    """
    AUTO = "auto"
    """
    Only the metadata that isn't represented by the template classes or excluded with ``meta_exclude`` will be included
    """
    NONE = None
    """
    No metadata will be included.
    """


class SplitMode(str, Enum):
    FULL = "full"
    """
    Import all classes defined in imported schemas
    """

    AUTO = "auto"
    """
    Only import those classes that are actually used in the generated schema as

    * parents (``is_a``)
    * mixins
    * slot ranges
    """


DefinitionType = TypeVar("DefinitionType", bound=Union[SchemaDefinition, ClassDefinition, SlotDefinition])
TemplateType = TypeVar("TemplateType", bound=Union[PydanticModule, PydanticClass, PydanticAttribute])


def make_valid_python_identifier(name: str) -> str:
    """
    Convert a string to a valid Python identifier.

    This is used when slot names contain characters that are not valid in Python
    identifiers (e.g., '@id', '@type'). The original name can be preserved using
    Pydantic field aliases.

    Args:
        name: The original name that may contain invalid characters

    Returns:
        A valid Python identifier that doesn't start with underscore (Pydantic restriction)
    """
    # Replace invalid characters with underscores
    identifier = re.sub(r"[^a-zA-Z0-9_]", "_", name)

    # Remove leading underscores (Pydantic doesn't allow field names starting with _)
    identifier = identifier.lstrip("_")

    # Ensure it doesn't start with a number
    if identifier and identifier[0].isdigit():
        identifier = f"field_{identifier}"

    # Ensure it's not a keyword
    if keyword.iskeyword(identifier):
        identifier = f"{identifier}_"

    # Ensure it's not empty
    if not identifier:
        identifier = "field"

    return identifier


@dataclass
class PydanticGenerator(OOCodeGenerator, LifecycleMixin):
    """
    Generates Pydantic-compliant classes from a schema

    This is an alternative to the dataclasses-based Pythongen

    Lifecycle methods (see :class:`.LifecycleMixin` ) supported:

    * :meth:`~.LifecycleMixin.before_generate_enums`

    Slot generation is nested within class generation, since the pydantic generator currently doesn't
    create an independent representation of slots aside from their materialization as class fields.
    Accordingly, the ``before_`` and ``after_generate_slots`` are called before and after each class's
    slot generation, rather than all slot generation.

    * :meth:`~.LifecycleMixin.before_generate_classes`
    * :meth:`~.LifecycleMixin.before_generate_class`
    * :meth:`~.LifecycleMixin.after_generate_class`
    * :meth:`~.LifecycleMixin.after_generate_classes`

    * :meth:`~.LifecycleMixin.before_generate_slots`
    * :meth:`~.LifecycleMixin.before_generate_slot`
    * :meth:`~.LifecycleMixin.after_generate_slot`
    * :meth:`~.LifecycleMixin.after_generate_slots`

    * :meth:`~.LifecycleMixin.before_render_template`
    * :meth:`~.LifecycleMixin.after_render_template`

    """

    # ClassVar overrides
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.2"
    valid_formats = ["pydantic"]
    file_extension = "py"

    # ObjectVars
    array_representations: list[ArrayRepresentation] = field(default_factory=lambda: [ArrayRepresentation.LIST])
    black: bool = False
    """
    If black is present in the environment, format the serialized code with it
    """

    template_dir: Optional[Union[str, Path]] = None
    """
    Override templates for each PydanticTemplateModel.

    Directory with templates that override the default :attr:`.PydanticTemplateModel.template`
    for each class. If a matching template is not found in the override directory,
    the default templates will be used.
    """
    extra_fields: Literal["allow", "forbid", "ignore"] = "forbid"
    gen_mixin_inheritance: bool = True
    injected_classes: Optional[list[Union[type, str]]] = None
    """
    A list/tuple of classes to inject into the generated module.

    Accepts either live classes or strings. Live classes will have their source code
    extracted with inspect.get - so they need to be standard python classes declared in a
    source file (ie. the module they are contained in needs a ``__file__`` attr,
    see: :func:`inspect.getsource` )
    """
    injected_fields: Optional[list[str]] = None
    """
    A list/tuple of field strings to inject into the base class.

    Examples:

    .. code-block:: python

        injected_fields = (
            'object_id: Optional[str] = Field(None, description="Unique UUID for each object")',
        )

    """
    imports: Optional[Union[list[Import], Imports]] = None
    """
    Additional imports to inject into generated module.

    Examples:

    .. code-block:: python

        from linkml.generators.pydanticgen.template import (
            ConditionalImport,
            ObjectImport,
            Import,
            Imports
        )

        imports = (Imports() +
            Import(module='sys') +
            Import(module='numpy', alias='np') +
            Import(module='pathlib', objects=[
                ObjectImport(name="Path"),
                ObjectImport(name="PurePath", alias="RenamedPurePath")
            ]) +
            ConditionalImport(
                module="typing",
                objects=[ObjectImport(name="Literal")],
                condition="sys.version_info >= (3, 8)",
                alternative=Import(
                    module="typing_extensions",
                    objects=[ObjectImport(name="Literal")]
                ),
            ).imports
        )

    becomes:

    .. code-block:: python

        import sys
        import numpy as np
        from pathlib import (
            Path,
            PurePath as RenamedPurePath
        )
        if sys.version_info >= (3, 8):
            from typing import Literal
        else:
            from typing_extensions import Literal

    """
    sort_imports: bool = True
    """
    Before returning from :meth:`.PydanticGenerator.render`, sort imports with :meth:`.Imports.sort`

    Default ``True``, but optional in case import order must be explicitly given,
    eg. to avoid circular import errors in complex generator subclasses.
    """
    metadata_mode: Union[MetadataMode, str, None] = MetadataMode.AUTO
    """
    How to include schema metadata in generated pydantic models.

    See :class:`.MetadataMode` for mode documentation
    """
    split: bool = False
    """
    Generate schema that import other schema as separate python modules
    that import from one another, rather than rolling all into a single
    module (default, ``False``).
    """
    split_pattern: str = ".{{ schema.name }}"
    """
    When splitting generation, imported modules need to be generated separately
    and placed in a python package and import from each other. Since the
    location of those imported modules is variable -- e.g. one might want to
    generate schema in multiple packages depending on their version -- this
    pattern is used to generate the module portion of the import statement.

    These patterns should generally yield a relative module import,
    since functions like :func:`.generate_split` will generate and write files
    relative to some base file, though this is not a requirement since custom
    split generation logic is also allowed.

    The pattern is a jinja template string that is given the ``SchemaDefinition``
    of the imported schema in the environment. Additional variables can be passed
    into the jinja environment with the :attr:`.split_context` argument.

    Further modification is possible by using jinja filters.

    After templating, the string is passed through a :attr:`SNAKE_CASE` pattern
    to replace whitespace and other characters that can't be used in module names.

    See also :meth:`.generate_module_import`, which is used to generate the
    module portion of the import statement (and can be overridden in subclasses).

    Examples:

        for a schema named ``ExampleSchema`` and version ``1.2.3`` ...

        ``".{{ schema.name }}"`` (the default) becomes

        ``from .example_schema import ClassA, ...``

        ``"...{{ schema.name }}.v{{ schema.version | replace('.', '_') }}"`` becomes

        ``from ...example_schema.v1_2_3 import ClassA, ...``

    """
    split_context: Optional[dict] = None
    """
    Additional variables to pass into ``split_pattern`` when
    generating imported module names.

    Passed in as ``**kwargs`` , so e.g. if ``split_context = {'myval': 1}``
    then one would use it in a template string like ``{{ myval }}``
    """
    split_mode: SplitMode = SplitMode.AUTO
    """
    How to filter imports from imported schema.

    See :class:`.SplitMode` for description of options
    """

    # ObjectVars (identical to pythongen)
    gen_classvars: bool = True
    gen_slots: bool = True
    genmeta: bool = False
    emit_metadata: bool = True

    # ClassVars
    SNAKE_CASE: ClassVar[str] = r"(((?<!^)(?<!\.))(?=[A-Z][a-z]))|([^\w\.]+)"
    """Substitute CamelCase and non-word characters with _"""

    # Private attributes
    _predefined_slot_values: Optional[dict[str, dict[str, str]]] = None
    _class_bases: Optional[dict[str, list[str]]] = None

    def __post_init__(self):
        super().__post_init__()

    def compile_module(self, **kwargs) -> ModuleType:
        """
        Compiles generated python code to a module
        :return:
        """
        pycode = self.serialize(**kwargs)
        try:
            return compile_python(pycode)
        except NameError as e:
            logger.error(f"Code:\n{pycode}")
            logger.error(f"Error compiling generated python code: {e}")
            raise e

    def _get_classes(self, sv: SchemaView) -> tuple[list[ClassDefinition], Optional[list[ClassDefinition]]]:
        all_classes = sv.all_classes(imports=True).values()

        if self.split:
            local_classes = sv.all_classes(imports=False).values()
            imported_classes = [c for c in all_classes if c not in local_classes]
            return list(local_classes), imported_classes
        else:
            return list(all_classes), None

    @staticmethod
    def sort_classes(
        clist: list[ClassDefinition], imported: Optional[list[ClassDefinition]] = None
    ) -> list[ClassDefinition]:
        """
        sort classes such that if C is a child of P then C appears after P in the list

        Overridden method include mixin classes

        TODO: This should move to SchemaView
        """
        if imported is not None:
            imported = [i.name for i in imported]

        clist = list(clist)
        slist = []  # sorted
        while len(clist) > 0:
            can_add = False
            for i in range(len(clist)):
                candidate = clist[i]
                can_add = False
                if candidate.is_a:
                    candidates = [candidate.is_a] + candidate.mixins
                else:
                    candidates = candidate.mixins

                # remove blocking classes imported from other schemas if in split mode
                if imported:
                    candidates = [c for c in candidates if c not in imported]

                if not candidates:
                    can_add = True
                else:
                    if set(candidates) <= set([p.name for p in slist]):
                        can_add = True
                if can_add:
                    slist = slist + [candidate]
                    del clist[i]
                    break
            if not can_add:
                raise ValueError(f"could not find suitable element in {clist} that does not ref {slist}")
        return slist

    def generate_class(self, cls: ClassDefinition) -> ClassResult:
        pyclass = PydanticClass(
            name=camelcase(cls.name),
            bases=self.class_bases.get(camelcase(cls.name), PydanticBaseModel.default_name),
            description=cls.description.replace('"', '\\"') if cls.description is not None else None,
        )

        imports = self._get_imports(cls) if self.split else None

        result = ClassResult(cls=pyclass, source=cls, imports=imports)

        # Gather slots
        slots = [self.schemaview.induced_slot(sn, cls.name) for sn in self.schemaview.class_slots(cls.name)]
        slots = self.before_generate_slots(slots, self.schemaview)

        slot_results = []
        for slot in slots:
            slot = self.before_generate_slot(slot, self.schemaview)
            slot = self.generate_slot(slot, cls)
            slot = self.after_generate_slot(slot, self.schemaview)
            slot_results.append(slot)
            result = result.merge(slot)

        slot_results = self.after_generate_slots(slot_results, self.schemaview)
        attributes = {slot.attribute.name: slot.attribute for slot in slot_results}

        result.cls.attributes = attributes
        result.cls = self.include_metadata(result.cls, cls)

        return result

    def generate_slot(self, slot: SlotDefinition, cls: ClassDefinition) -> SlotResult:
        slot_args = {
            k: getattr(slot, k, None)
            for k in PydanticAttribute.model_fields.keys()
            if getattr(slot, k, None) is not None
        }
        slot_alias = slot.alias if slot.alias else slot.name

        # Create a valid Python identifier for the field name
        python_field_name = make_valid_python_identifier(underscore(slot_alias))
        slot_args["name"] = python_field_name

        # If the original name is different from the Python identifier, set an alias
        if slot_alias != python_field_name:
            slot_args["alias"] = slot_alias
        else:
            # Remove any existing alias if the names are the same
            if "alias" in slot_args:
                del slot_args["alias"]

        slot_args["description"] = slot.description.replace('"', '\\"') if slot.description is not None else None
        predef = self.predefined_slot_values.get(camelcase(cls.name), {}).get(slot.name, None)
        if predef is not None:
            slot_args["predefined"] = str(predef)

        pyslot = PydanticAttribute(**slot_args)
        pyslot = self.include_metadata(pyslot, slot)

        slot_ranges = []
        # Confirm that the original slot range (ignoring the default that comes in from
        # induced_slot) isn't in addition to setting any_of
        any_of_ranges = [a.range if a.range else slot.range for a in slot.any_of]
        if any_of_ranges:
            # list comprehension here is pulling ranges from within AnonymousSlotExpression
            slot_ranges.extend(any_of_ranges)
        else:
            slot_ranges.append(slot.range)

        pyranges = [self.generate_python_range(slot_range, slot, cls) for slot_range in slot_ranges]

        pyranges = list(set(pyranges))  # remove duplicates
        pyranges.sort()

        if len(pyranges) == 1:
            pyrange = pyranges[0]
        elif len(pyranges) > 1:
            pyrange = f"Union[{', '.join(pyranges)}]"
        else:
            raise Exception(f"Could not generate python range for {cls.name}.{slot.name}")

        pyslot.range = pyrange

        imports = self._get_imports(slot) if self.split else None

        result = SlotResult(attribute=pyslot, source=slot, imports=imports)

        if slot.array is not None:
            results = self.get_array_representations_range(slot, result.attribute.range)
            if len(results) == 1:
                result.attribute.range = results[0].range
            else:
                result.attribute.range = f"Union[{', '.join([res.range for res in results])}]"
            for res in results:
                result = result.merge(res)

        elif slot.multivalued:
            if slot.inlined or slot.inlined_as_list:
                collection_key = self.generate_collection_key(slot_ranges, slot, cls)
            else:
                collection_key = None
            if slot.inlined is False or collection_key is None or slot.inlined_as_list is True:
                result.attribute.range = f"list[{result.attribute.range}]"
            else:
                simple_dict_value = None
                if len(slot_ranges) == 1:
                    simple_dict_value = self._inline_as_simple_dict_with_value(slot)
                if simple_dict_value:
                    # simple_dict_value might be the range of the identifier of a class when range is a class,
                    # so we specify either that identifier or the range itself
                    if simple_dict_value != result.attribute.range:
                        simple_dict_value = f"Union[{simple_dict_value}, {result.attribute.range}]"
                    result.attribute.range = f"dict[str, {simple_dict_value}]"
                else:
                    result.attribute.range = f"dict[{collection_key}, {result.attribute.range}]"
        if not (slot.required or slot.identifier or slot.key) and not slot.designates_type:
            result.attribute.range = f"Optional[{result.attribute.range}]"
        return result

    @property
    def predefined_slot_values(self) -> dict[str, dict[str, str]]:
        """
        :return: Dictionary of dictionaries with predefined slot values for each class
        """
        if self._predefined_slot_values is None:
            sv = self.schemaview
            ifabsent_processor = PydanticIfAbsentProcessor(sv)
            slot_values = defaultdict(dict)
            for class_def in sv.all_classes().values():
                for slot_name in sv.class_slots(class_def.name):
                    slot = sv.induced_slot(slot_name, class_def.name)
                    if slot.designates_type:
                        target_value = get_type_designator_value(sv, slot, class_def)
                        slot_values[camelcase(class_def.name)][slot.name] = f'"{target_value}"'
                        if slot.multivalued:
                            slot_values[camelcase(class_def.name)][slot.name] = (
                                "[" + slot_values[camelcase(class_def.name)][slot.name] + "]"
                            )
                        slot_values[camelcase(class_def.name)][slot.name] = slot_values[camelcase(class_def.name)][
                            slot.name
                        ]
                    elif slot.ifabsent is not None:
                        value = ifabsent_processor.process_slot(slot, class_def)
                        slot_values[camelcase(class_def.name)][slot.name] = value

                self._predefined_slot_values = slot_values

        return self._predefined_slot_values

    @property
    def class_bases(self) -> dict[str, list[str]]:
        """
        Generate the inheritance list for each class from is_a plus mixins
        :return:
        """
        if self._class_bases is None:
            sv = self.schemaview
            parents = {}
            for class_def in sv.all_classes().values():
                class_parents = []
                if class_def.is_a:
                    class_parents.append(camelcase(class_def.is_a))
                if self.gen_mixin_inheritance and class_def.mixins:
                    class_parents.extend([camelcase(mixin) for mixin in class_def.mixins])
                if len(class_parents) > 0:
                    # Use the sorted list of classes to order the parent classes, but reversed to match MRO needs
                    class_parents.sort(
                        key=lambda x: self.sorted_class_names.index(x) if x in self.sorted_class_names else -1
                    )
                    class_parents.reverse()
                    parents[camelcase(class_def.name)] = class_parents
            self._class_bases = parents
        return self._class_bases

    def get_mixin_identifier_range(self, mixin) -> str:
        sv = self.schemaview
        id_ranges = list(
            {
                _get_pyrange(sv.get_type(sv.get_identifier_slot(c).range), sv)
                for c in sv.class_descendants(mixin.name, mixins=True)
                if sv.get_identifier_slot(c) is not None
            }
        )
        if len(id_ranges) == 0:
            return None
        elif len(id_ranges) == 1:
            return id_ranges[0]
        else:
            return f"Union[{'.'.join(id_ranges)}]"

    def get_class_slot_range(self, slot_range: str, inlined: bool, inlined_as_list: bool) -> str:
        sv = self.schemaview
        range_cls = sv.get_class(slot_range)

        # Hardcoded handling for Any
        if range_cls.class_uri == "linkml:Any":
            return "Any"

        # Inline the class itself only if the class is defined as inline, or if the class has no
        # identifier slot and also isn't a mixin.
        if (
            inlined
            or inlined_as_list
            or (sv.get_identifier_slot(range_cls.name, use_key=True) is None and not sv.is_mixin(range_cls.name))
        ):
            if (
                len([x for x in sv.class_induced_slots(slot_range) if x.designates_type]) > 0
                and len(sv.class_descendants(slot_range)) > 1
            ):
                return "Union[" + ",".join([camelcase(c) for c in sv.class_descendants(slot_range)]) + "]"
            else:
                return f"{camelcase(slot_range)}"

        # For the more difficult cases, set string as the default and attempt to improve it
        range_cls_identifier_slot_range = "str"

        # For mixins, try to use the identifier slot of descendant classes
        if self.gen_mixin_inheritance and sv.is_mixin(range_cls.name) and sv.get_identifier_slot(range_cls.name):
            range_cls_identifier_slot_range = self.get_mixin_identifier_range(range_cls)

        # If the class itself has an identifier slot, it can be allowed to overwrite a value from mixin above
        if (
            sv.get_identifier_slot(range_cls.name) is not None
            and sv.get_identifier_slot(range_cls.name).range is not None
        ):
            range_cls_identifier_slot_range = _get_pyrange(
                sv.get_type(sv.get_identifier_slot(range_cls.name).range), sv
            )

        return range_cls_identifier_slot_range

    def generate_python_range(self, slot_range, slot_def: SlotDefinition, class_def: ClassDefinition) -> str:
        """
        Generate the python range for a slot range value
        """
        sv = self.schemaview

        if slot_def.designates_type:
            pyrange = (
                "Literal["
                + ",".join(['"' + x + '"' for x in get_accepted_type_designator_values(sv, slot_def, class_def)])
                + "]"
            )
        elif slot_def.equals_string:
            pyrange = f'Literal["{slot_def.equals_string}"]'
        elif slot_def.equals_string_in:
            pyrange = "Literal[" + ", ".join([f'"{a_string}"' for a_string in slot_def.equals_string_in]) + "]"
        elif slot_range in sv.all_classes():
            pyrange = self.get_class_slot_range(
                slot_range,
                inlined=slot_def.inlined,
                inlined_as_list=slot_def.inlined_as_list,
            )
        elif slot_range in sv.all_enums():
            pyrange = f"{camelcase(slot_range)}"
        elif slot_range in sv.all_types():
            t = sv.get_type(slot_range)
            pyrange = _get_pyrange(t, sv)
        elif slot_range is None:
            pyrange = "str"
        else:
            # TODO: default ranges in schemagen
            # pyrange = 'str'
            # logger.error(f'range: {s.range} is unknown')
            raise Exception(f"range: {slot_range}")
        return pyrange

    def generate_collection_key(
        self,
        slot_ranges: list[str],
        slot_def: SlotDefinition,
        class_def: ClassDefinition,
    ) -> Optional[str]:
        """
        Find the python range value (str, int, etc) for the identifier slot
        of a class used as a slot range.

        If a pyrange value matches a class name, the range of the identifier slot
        will be returned. If more than one match is found and they don't match,
        an exception will be raised.

        :param slot_ranges: list of python range values
        """

        collection_keys: set[str] = set()

        if slot_ranges is None:
            return None

        for slot_range in slot_ranges:
            if slot_range is None or slot_range not in self.schemaview.all_classes():
                continue  # ignore non-class ranges

            identifier_slot = self.schemaview.get_identifier_slot(slot_range, use_key=True)
            if identifier_slot is not None:
                collection_keys.add(self.generate_python_range(identifier_slot.range, slot_def, class_def))
        if len(collection_keys) > 1:
            raise Exception(f"Slot with any_of range has multiple identifier slot range types: {collection_keys}")
        if len(collection_keys) == 1:
            return list(collection_keys)[0]
        return None

    def _clean_injected_classes(self, injected_classes: list[Union[str, type]]) -> Optional[list[str]]:
        """Get source, deduplicate, and dedent injected classes"""
        if len(injected_classes) == 0:
            return None

        injected_classes = list(
            dict.fromkeys([c if isinstance(c, str) else inspect.getsource(c) for c in injected_classes])
        )
        injected_classes = [textwrap.dedent(c) for c in injected_classes]
        return injected_classes

    def _inline_as_simple_dict_with_value(self, slot_def: SlotDefinition) -> Optional[str]:
        """
        Determine if a slot should be inlined as a simple dict with a value.

        For example, if we have a class such as Prefix, with two slots prefix and expansion,
        then an inlined list of prefixes can be serialized as:

        .. code-block:: yaml

            prefixes:
                prefix1: expansion1
                prefix2: expansion2
                ...

        Provided that the prefix slot is the identifier slot for the Prefix class.

        TODO: move this to SchemaView

        :param slot_def: SlotDefinition
        :param sv: SchemaView
        :return: str
        """
        if slot_def.inlined and not slot_def.inlined_as_list:
            if slot_def.range in self.schemaview.all_classes():
                id_slot = self.schemaview.get_identifier_slot(slot_def.range, use_key=True)
                if id_slot is not None:
                    range_cls_slots = self.schemaview.class_induced_slots(slot_def.range)
                    if len(range_cls_slots) == 2:
                        non_id_slots = [slot for slot in range_cls_slots if slot.name != id_slot.name]
                        if len(non_id_slots) == 1:
                            value_slot = non_id_slots[0]
                            value_slot_range_type = self.schemaview.get_type(value_slot.range)
                            if value_slot_range_type is not None:
                                return _get_pyrange(value_slot_range_type, self.schemaview)
        return None

    def _template_environment(self) -> Environment:
        env = PydanticTemplateModel.environment()
        if self.template_dir is not None:
            loader = ChoiceLoader([FileSystemLoader(self.template_dir), env.loader])
            env.loader = loader
        return env

    def get_array_representations_range(self, slot: SlotDefinition, range: str) -> list[SlotResult]:
        """
        Generate the python range for array representations
        """
        array_reps = []
        for repr in self.array_representations:
            generator = ArrayRangeGenerator.get_generator(repr)
            result = generator(slot.array, range).make()
            array_reps.append(result)

        if len(array_reps) == 0:
            raise ValueError("No array representation generated, but one was requested!")

        return array_reps

    @overload
    def include_metadata(self, model: PydanticModule, source: SchemaDefinition) -> PydanticModule: ...

    @overload
    def include_metadata(self, model: PydanticClass, source: ClassDefinition) -> PydanticClass: ...

    @overload
    def include_metadata(self, model: PydanticAttribute, source: SlotDefinition) -> PydanticAttribute: ...

    def include_metadata(self, model: TemplateType, source: DefinitionType) -> TemplateType:
        """
        Include metadata from the source schema that is otherwise not represented in the pydantic template models.

        Metadata inclusion mode is dependent on :attr:`.metadata_mode` - see:

        - :class:`.MetadataMode`
        - :meth:`.PydanticTemplateModel.exclude_from_meta`

        """
        if self.metadata_mode is None or self.metadata_mode == MetadataMode.NONE:
            return model
        elif self.metadata_mode in (MetadataMode.AUTO, MetadataMode.AUTO.value):
            meta = {k: v for k, v in remove_empty_items(source).items() if k not in model.exclude_from_meta()}
        elif self.metadata_mode in (MetadataMode.EXCEPT_CHILDREN, MetadataMode.EXCEPT_CHILDREN.value):
            meta = {}
            for k, v in remove_empty_items(source).items():
                if k in ("slots", "classes") and isinstance(model, PydanticModule):
                    # FIXME: Special-case removal of slots until we generate class-level slots
                    continue

                if not hasattr(model, k):
                    meta[k] = v
                    continue

                model_attr = getattr(model, k)
                if isinstance(model_attr, list) and not any(
                    [isinstance(item, PydanticTemplateModel) for item in model_attr]
                ):
                    meta[k] = v
                elif isinstance(model_attr, dict) and not any(
                    [isinstance(item, PydanticTemplateModel) for item in model_attr.values()]
                ):
                    meta[k] = v
                elif not isinstance(model_attr, (list, dict, PydanticTemplateModel)):
                    meta[k] = v

        elif self.metadata_mode in (MetadataMode.FULL, MetadataMode.FULL.value):
            meta = remove_empty_items(source)
        else:
            raise ValueError(
                f"Unknown metadata mode '{self.metadata_mode}', needs to be one of {[mode for mode in MetadataMode]}"
            )

        model.meta = meta
        return model

    def _get_imports(self, element: Union[ClassDefinition, SlotDefinition, None] = None) -> Imports:
        """
        Get imports that are implied by their usage in slots or classes
        (and thus need to be imported when generating schemas in :attr:`.split` == ``True`` mode).

        **Note:**
        Since in pydantic (currently) the only things that are materialized are classes, we don't
        import class slots from imported schemas and abandon slots, directly expressing them
        in the model.

        This is a parent placeholder method in case that changes, "give me something and return
        a set of imports" that calls subordinate methods. If slots become materialized, keep
        this as the directly called method rather than spaghetti-ing out another
        independent method. This method is also isolated in anticipation of structured imports,
        where we will need to revise our expectations of what is imported when.

        Args:
            element (:class:`.ClassDefinition` , :class:`.SlotDefinition` , None): The element
                to get import for. If ``None`` , get all needed imports (see :attr:`.split_mode`
        """
        # import from local references, rather than serializing every class in every file
        if not self.split or (self.split_mode == SplitMode.FULL and element is not None):
            # we are either compiling this whole thing in one big file (default)
            # or going to import all classes from the imported schemas,
            # so we don't import anything
            return Imports()

        # gather a list of class names,
        # remove local classes and transform to Imports later.
        needed_classes = []

        # fine to call rather than pass bc it's cached
        all_classes = self.schemaview.all_classes(imports=True)
        local_classes = self.schemaview.all_classes(imports=False)

        if isinstance(element, ClassDefinition):
            if element.is_a:
                needed_classes.append(element.is_a)
            if element.mixins:
                needed_classes.extend(element.mixins)

        elif isinstance(element, SlotDefinition):
            # collapses `slot.range`, `slot.any_of`, and `slot.one_of` to a list
            slot_ranges = self.schemaview.slot_range_as_union(element)
            needed_classes.extend([a_range for a_range in slot_ranges if a_range in all_classes])

        elif element is None:
            # get all imports
            needed_classes.extend([cls for cls in all_classes if cls not in local_classes])

        else:
            raise ValueError(f"Unsupported type of element to get imports from: f{type(element)}")

        # SPECIAL CASE: classes that are not generated for structural reasons.
        # TODO: Do we want to have a general means of skipping class generation?
        skips = ("AnyType",)

        class_imports = [
            self._get_element_import(cls) for cls in needed_classes if (cls not in local_classes and cls not in skips)
        ]
        imports = Imports(imports=class_imports)

        return imports

    def generate_module_import(self, schema: SchemaDefinition, context: Optional[dict] = None) -> str:
        """
        Generate the module string for importing from python modules generated from imported schemas
        when in :attr:`.split` mode.

        Use the :attr:`.split_pattern` as a jinja template rendered with the :class:`.SchemaDefinition`
        and any passed ``context``. Apply the :attr:`.SNAKE_CASE` regex to substitute matches with
        ``_`` and ensure lowercase.
        """
        if context is None:
            context = {}
        module = Template(self.split_pattern).render(schema=schema, **context)
        module = re.sub(self.SNAKE_CASE, "_", module) if self.SNAKE_CASE else module
        module = module.lower()
        return module

    def _get_element_import(self, class_name: ElementName) -> Import:
        """
        Make an import object for an element from another schema, using the
        :attr:`.split_import_pattern` to generate the module import part.
        """
        schema_name = self.schemaview.element_by_schema_map()[class_name]
        schema = [s for s in self.schemaview.schema_map.values() if s.name == schema_name][0]
        module = self.generate_module_import(schema, self.split_context)
        return Import(module=module, objects=[ObjectImport(name=camelcase(class_name))], is_schema=True)

    def render(self) -> PydanticModule:
        """
        Render the schema to a :class:`PydanticModule` model
        """
        sv: SchemaView
        sv = self.schemaview

        # imports
        imports = DEFAULT_IMPORTS
        if self.imports is not None:
            if isinstance(self.imports, Imports):
                imports += self.imports
            else:
                for i in self.imports:
                    imports += i
        if self.split_mode == SplitMode.FULL:
            imports += self._get_imports()

        # injected classes
        injected_classes = DEFAULT_INJECTS.copy()
        if self.injected_classes is not None:
            injected_classes += self.injected_classes.copy()

        # enums
        enums = self.before_generate_enums(list(sv.all_enums().values()), sv)
        enums = self.generate_enums({e.name: e for e in enums})

        base_model = PydanticBaseModel(extra_fields=self.extra_fields, fields=self.injected_fields)

        # schema classes
        class_results = []
        source_classes, imported_classes = self._get_classes(sv)
        source_classes = self.sort_classes(source_classes, imported_classes)
        # Don't want to generate classes when class_uri is linkml:Any, will
        # just swap in typing.Any instead down below
        source_classes = [c for c in source_classes if c.class_uri != "linkml:Any"]
        source_classes = self.before_generate_classes(source_classes, sv)
        self.sorted_class_names = [camelcase(c.name) for c in source_classes]
        for cls in source_classes:
            cls = self.before_generate_class(cls, sv)
            result = self.generate_class(cls)
            result = self.after_generate_class(result, sv)
            class_results.append(result)
            if result.imports is not None:
                imports += result.imports
            if result.injected_classes is not None:
                injected_classes.extend(result.injected_classes)

        class_results = self.after_generate_classes(class_results, sv)

        classes = {r.cls.name: r.cls for r in class_results}
        injected_classes = self._clean_injected_classes(injected_classes)

        imports.render_sorted = self.sort_imports

        module = PydanticModule(
            metamodel_version=self.schema.metamodel_version,
            version=self.schema.version,
            python_imports=imports,
            base_model=base_model,
            injected_classes=injected_classes,
            enums=enums,
            classes=classes,
        )
        module = self.include_metadata(module, self.schemaview.schema)
        module = self.before_render_template(module, self.schemaview)
        return module

    def serialize(self, rendered_module: Optional[PydanticModule] = None) -> str:
        """
        Serialize the schema to a pydantic module as a string

        Args:
            rendered_module ( :class:`.PydanticModule` ): Optional, if schema was previously
                rendered with :meth:`~.PydanticGenerator.render` , use that,
                otherwise :meth:`~.PydanticGenerator.render` fresh.
        """
        if rendered_module is not None:
            module = rendered_module
        else:
            module = self.render()
        serialized = module.render(self._template_environment(), self.black)
        serialized = self.after_render_template(serialized, self.schemaview)
        return serialized

    def default_value_for_type(self, typ: str) -> str:
        return "None"

    @classmethod
    def generate_split(
        cls,
        schema: Union[str, Path, SchemaDefinition],
        output_path: Union[str, Path] = Path("."),
        split_pattern: Optional[str] = None,
        split_context: Optional[dict] = None,
        split_mode: SplitMode = SplitMode.AUTO,
        **kwargs,
    ) -> list[SplitResult]:
        """
        Generate a schema that imports from other schema as a set of python modules that
        import from one another, rather than generating all imported classes in a single schema.

        Uses ``output_path`` for the main schema from ``schema`` , and then
        generates any imported schema (from which classes are actually used)
        to modules whose locations are determined by the module names generated
        by the ``split_pattern`` (see :attr:`.PydanticGenerator.split_pattern` ).

        For example, for

        * a ``output_path`` of ``my_dir/v1_2_3/main.py``
        * a schema ``main`` with a version ``v1.2.3``
        * that imports from ``s2`` with version ``v4.5.6``,
        * and a ``split_pattern`` of ``..{{ schema.version | replace('.', '_') }}.{{ schema.name }}``

        One would get:
        * ``my_dir/v1_2_3/main.py`` , as expected
        * that imports ``from ..v4_5_6.s2``
        * a module at ``my_dir/v4_5_6/s2.py``

        ``__init__.py`` files are generated for any directories that are between
        the generated modules and their highest common directory.

        Args:
            schema (str, :class:`.Path` , :class:`.SchemaDefinition` ): Main schema to generate
            output_path (str, :class:`.Path` ): Python ``.py`` module to generate main schema to
            split_pattern (str): Pattern to use to generate module names, see :attr:`.PydanticGenerator.split_pattern`
            split_context (dict): Additional variables to pass into jinja context when generating module import names.

        Returns:
            list[:class:`.SplitResult`]
        """
        output_path = Path(output_path)
        if not output_path.suffix == ".py":
            raise ValueError(f"output path must be a python file to write the main schema to, got {output_path}")

        results = []

        # --------------------------------------------------
        # Main schema
        # --------------------------------------------------
        gen_kwargs = kwargs
        gen_kwargs.update(
            {"split": True, "split_pattern": split_pattern, "split_context": split_context, "split_mode": split_mode}
        )
        generator = cls(schema, **gen_kwargs)
        # Generate the initial schema to figure out which of the imported schema actually need
        # to be generated
        rendered = generator.render()
        # write schema - we use the ``output_path`` for the main schema, and then
        # interpret all imported schema paths as relative to that
        output_path.parent.mkdir(parents=True, exist_ok=True)
        serialized = generator.serialize(rendered_module=rendered)
        with open(output_path, "w", encoding="utf-8") as ofile:
            ofile.write(serialized)

        results.append(
            SplitResult(main=True, source=generator.schemaview.schema, path=output_path, serialized_module=serialized)
        )

        # --------------------------------------------------
        # Imported schemas
        # --------------------------------------------------
        imported_schema = {
            generator.generate_module_import(sch): sch for sch in generator.schemaview.schema_map.values()
        }
        for generated_import in [i for i in rendered.python_imports if i.is_schema]:
            import_generator = cls(imported_schema[generated_import.module], **gen_kwargs)
            serialized = import_generator.serialize()
            rel_path = _import_to_path(generated_import.module)
            abs_path = (output_path.parent / rel_path).resolve()
            abs_path.parent.mkdir(parents=True, exist_ok=True)
            with open(abs_path, "w", encoding="utf-8") as ofile:
                ofile.write(serialized)

            results.append(
                SplitResult(
                    main=False,
                    source=imported_schema[generated_import.module],
                    path=abs_path,
                    serialized_module=serialized,
                    module_import=generated_import.module,
                )
            )

        _ensure_inits([r.path for r in results])
        return results


def _subclasses(cls: type):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in _subclasses(c)])


_TEMPLATE_NAMES = sorted(list(set([c.template for c in _subclasses(PydanticTemplateModel)])))


def _import_to_path(module: str) -> Path:
    """Make a (relative) ``Path`` object from a python module import string"""
    # handle leading .'s separately..
    _, dots, module = re.split(r"(^\.*)(?=\w)", module, maxsplit=1)
    # treat zero or one dots as a relative import to the current directory
    dir_pieces = ["../" for _ in range(max(len(dots) - 1, 0))]
    dir_pieces.extend(module.split("."))
    dir_pieces[-1] = dir_pieces[-1] + ".py"
    return Path(*dir_pieces)


def _ensure_inits(paths: list[Path]):
    """For a set of paths, find the common root and it and all the subdirectories have an __init__.py"""
    # if there is only one file, there is no relative importing to be done
    if len(paths) <= 1:
        return
    common_path = Path(os.path.commonpath(paths))

    if not (ipath := (common_path / "__init__.py")).exists():
        with open(ipath, "w", encoding="utf-8") as ifile:
            ifile.write(" \n")

    for path in paths:
        # ensure __init__ for each directory from this path up to the common path
        path = path.parent
        while path != common_path:
            if not (ipath := (path / "__init__.py")).exists():
                with open(ipath, "w", encoding="utf-8") as ifile:
                    ifile.write(" \n")
            path = path.parent


@shared_arguments(PydanticGenerator)
@click.option("--template-file", hidden=True)
@click.option(
    "--template-dir",
    type=click.Path(),
    help="""
Optional jinja2 template directory to use for class generation.

Pass a directory containing templates with the same name as any of the default
:class:`.PydanticTemplateModel` templates to override them. The given directory will be
searched for matching templates, and use the default templates as a fallback
if an override is not found

Available templates to override:

\b
"""
    + "\n".join(["- " + name for name in _TEMPLATE_NAMES]),
)
@click.option(
    "--array-representations",
    type=click.Choice([k.value for k in ArrayRepresentation]),
    multiple=True,
    default=["list"],
    help="List of array representations to accept for array slots. Default is list of lists.",
)
@click.option(
    "--extra-fields",
    type=click.Choice(["allow", "ignore", "forbid"], case_sensitive=False),
    default="forbid",
    help="How to handle extra fields in BaseModel.",
)
@click.option(
    "--black",
    is_flag=True,
    default=False,
    help="Format generated models with black (must be present in the environment)",
)
@click.option(
    "--meta",
    type=click.Choice([k.value for k in MetadataMode]),
    default="auto",
    help="How to include linkml schema metadata in generated pydantic classes. "
    "See docs for MetadataMode for full description of choices. "
    "Default (auto) is to include all metadata that can't be otherwise represented",
)
@click.version_option(__version__, "-V", "--version")
@click.command(name="pydantic")
def cli(
    yamlfile,
    template_file=None,
    template_dir: Optional[str] = None,
    head=True,
    genmeta=False,
    classvars=True,
    slots=True,
    array_representations=list("list"),
    extra_fields: Literal["allow", "forbid", "ignore"] = "forbid",
    black: bool = False,
    meta: MetadataMode = "auto",
    **args,
):
    """Generate pydantic classes to represent a LinkML model"""
    if template_file is not None:
        raise DeprecationWarning(
            "Passing a single template_file is deprecated. Pass a directory of template files instead. "
            "See help string for --template-dir"
        )

    if template_dir is not None:
        if not Path(template_dir).exists():
            raise FileNotFoundError(f"The template directory {template_dir} does not exist!")

    gen = PydanticGenerator(
        yamlfile,
        array_representations=[ArrayRepresentation(x) for x in array_representations],
        extra_fields=extra_fields,
        emit_metadata=head,
        genmeta=genmeta,
        gen_classvars=classvars,
        gen_slots=slots,
        template_dir=template_dir,
        black=black,
        metadata_mode=meta,
        **args,
    )
    print(gen.serialize(), end="")


if __name__ == "__main__":
    cli()



"""
Generate dotfiles
"""

import os
from dataclasses import dataclass
from typing import Optional

import click
from deprecated.classic import deprecated
from graphviz import FORMATS, Digraph
from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import underscore

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

valid_formats = sorted(list(FORMATS))


@deprecated("Replaced by yuml/mermaid")
@dataclass
class DotGenerator(Generator):
    """
    Generates dotfiles

    No longer in use: use mermaid generator instead
    """

    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    directory_output = True
    valid_formats = ["png"] + valid_formats
    visit_all_class_slots = True
    uses_schemaloader = True

    # ObjectVars
    classnames: Optional[list[str]] = None
    filename: Optional[str] = None
    dirname: Optional[str] = None
    filedot: Optional[Digraph] = None
    classdot: Optional[Digraph] = None
    cls_subj: Optional[SlotDefinition] = None
    cls_obj: Optional[SlotDefinition] = None
    classname: Optional[list[str]] = None
    directory: Optional[str] = None

    def visit_schema(
        self,
        classname: Optional[list[str]] = None,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        **_,
    ) -> None:
        self.classnames = [] if classname is None else list(classname)
        for classname in self.classnames:
            if classname not in self.schema.classes:
                raise ValueError(f"Unknown class name: {classname}")
        self.filename = filename
        self.dirname = directory
        if filename:
            self.filedot = Digraph(comment=self.schema.name)
        if directory:
            os.makedirs(directory, exist_ok=True)

    def end_schema(self, **_) -> None:
        if self.filedot:
            self.filedot.render(
                self.filename,
                self.dirname,
                view=False,
                cleanup=True,
                format=self.format,
            )

    def visit_class(self, cls: ClassDefinition) -> bool:
        if self.classnames and cls.name not in self.classnames:
            return False
        if self.dirname:
            self.classdot = Digraph(comment=self.schema.name)
        self.node(cls.name, cls.name)
        if cls.is_a:
            self.edge(cls.name, cls.is_a, label="is_a")
        for mixin in cls.mixins:
            self.edge(cls.name, mixin, label="uses")
        self.cls_subj = self.cls_obj = None
        return True

    def end_class(self, cls: ClassDefinition) -> None:
        if self.cls_subj and self.cls_obj:
            rnode = "relation"
            self.edge(
                self.aliased_slot_name(self.cls_subj),
                self.aliased_slot_name(self.cls_obj),
                label=rnode,
            )
            self.edge(self.aliased_slot_name(self.cls_subj), rnode, style="dotted")
            self.edge(self.aliased_slot_name(self.cls_obj), rnode, style="dotted")
        if self.classdot:
            self.classdot.render(
                underscore(cls.name),
                self.dirname,
                view=False,
                cleanup=True,
                format=self.format,
            )

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition):
        if aliased_slot_name == "subject":
            self.cls_subj = slot
        elif aliased_slot_name == "object":
            self.cls_obj = slot
        color = "blue" if slot.name in cls.slots else "black"
        style = "dashed" if slot.alias in cls.slots else "solid"
        self.edge(
            cls.name,
            aliased_slot_name,
            label=aliased_slot_name,
            color=color,
            style=style,
        )
        srange = slot.range if slot.range else "Thing"
        self.node(slot.name, srange, color=color)

    def node(self, *args, **kwargs) -> None:
        if self.classdot:
            self.classdot.node(*args, **kwargs)
        if self.filedot:
            self.filedot.node(*args, **kwargs)

    def edge(self, *args, **kwargs) -> None:
        if self.classdot:
            self.classdot.edge(*args, **kwargs)
        if self.filedot:
            self.filedot.edge(*args, **kwargs)


@shared_arguments(DotGenerator)
@click.command(name="graphviz")
@click.option(
    "--directory",
    "-d",
    help="Output directory - if supplied, a graph per class will be generated",
)
@click.option("--out", "-o", help="Target file -- if supplied, one large graph will be generated")
@click.option("--classname", "-c", multiple=True, help="Class(es) to transform")
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, out, **args):
    """Generate graphviz representations of the LinkML model"""
    DotGenerator(yamlfile, **args).serialize(filename=out, **args)


if __name__ == "__main__":
    cli()



from linkml_runtime.linkml_model import (
    ClassDefinition,
    EnumDefinitionName,
    SlotDefinition,
)

from linkml.generators.python.python_ifabsent_processor import PythonIfAbsentProcessor


class PydanticIfAbsentProcessor(PythonIfAbsentProcessor):
    def map_enum_default_value(
        self, enum_name: EnumDefinitionName, permissible_value_name: str, slot: SlotDefinition, cls: ClassDefinition
    ):
        return f"{enum_name}.{permissible_value_name}"
